#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library item widget implementation
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

from tpDcc.libs import qt
from tpDcc.libs.qt.widgets.library import consts, mixin, items


class LibraryTreeWidget(mixin.LibraryViewWidgetMixin, QTreeWidget):
    """
    Class that implemented library tree viewer widget
    This class is used by LibraryViewer class
    """

    def __init__(self, parent=None):
        QTreeWidget.__init__(self, parent)
        mixin.LibraryViewWidgetMixin.__init__(self)

        self._header_labels = list()
        self._hidden_columns = dict()

        self.setAutoScroll(False)
        self.setMouseTracking(True)
        self.setSortingEnabled(False)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)

        header = self.header()
        header.setContextMenuPolicy(Qt.CustomContextMenu)
        header.customContextMenuRequested.connect(self._on_show_header_menu)

        self.itemClicked.connect(self._on_item_clicked)
        self.itemDoubleClicked.connect(self._on_item_double_clicked)

    """
    ##########################################################################################
    MIXIN
    ##########################################################################################
    """

    def _on_now(self, *args, **kwargs):
        print('asdfasfasfasdfasdfasdf')

    def mouseMoveEvent(self, event):
        """
        Triggered when the user moves the mouse over the current viewport
        :param event: QMouseEvent
        """

        mixin.LibraryViewWidgetMixin.mouseMoveEvent(self, event)
        QTreeWidget.mouseMoveEvent(self, event)

    def mouseReleaseEvent(self, event):
        """
        Triggerd when the user releases the mouse button on the viewport
        :param event: QMouseEvent
        """

        mixin.LibraryViewWidgetMixin.mouseReleaseEvent(self, event)
        QTreeWidget.mouseReleaseEvent(self, event)

    """
    ##########################################################################################
    BASE
    ##########################################################################################
    """

    def drawRow(self, painter, options, index):
        """
        Overrides base QTreeWidget drawDrow function
        :param painter: QPainter
        :param options: QStyleOption
        :param index: QModelIndex
        """

        item = self.itemFromIndex(index)
        item.paint_row(painter, options, index)

    def setColumnHidden(self, column, value):
        """
        Overrides base QTreeWidget setColumnHidden function
        :param column: int or str
        :param value: bool
        """

        if isinstance(column, (unicode, str)):
            column = self.column_from_label(column)

        label = self.label_from_column(column)
        self._hidden_columns[label] = value

        super(LibraryTreeWidget, self).setColumnHidden(column, value)

        width = self.columnWidth(column)
        if width < consts.TREE_MINIMUM_WIDTH:
            width = consts.TREE_DEFAULT_WIDTH
            self.setColumnWidth(column, width)

    def resizeColumnToContents(self, column):
        """
        Overrides base QTreeWidget resizeColumnToContents function
        Resize the given column to the data of that column
        :param column: int or str
        """

        width = 0
        for item in self.items():
            text = item.text(column)
            font = item.font(column)
            metrics = QFontMetrics(font)
            text_width = metrics.width(text) + item.padding()
            width = max(width, text_width)

        self.setColumnWidth(column, width)

    def setHeaderLabels(self, labels):
        """
        Overrides base QTreeWidget setHeaderLabels function
        :param labels: list(str)
        """

        labels = self._remove_duplicates(labels)
        column_settings = self.column_settings()
        super(LibraryTreeWidget, self).setHeaderLabels(labels)
        self._header_labels = labels
        self.update_column_hidden()
        self.set_column_settings(column_settings)

    def items(self):
        """
        Overrides base QTreeWidget items function
        Return a list of all items in the tree widget
        :return: list(LibraryItem)
        """

        items_list = list()
        for item in self._items():
            if not isinstance(item, items.LibraryGroupItem):
                items_list.append(item)

        return items_list

    def selectedItems(self):
        """
        Overrides base QTreeWidget selectedItems function
        Returns all selected items
        :return: list(LibraryItem)
        """

        items_list = list()
        items_ = super(LibraryTreeWidget, self).selectedItems()

        for item in items_:
            if not isinstance(item, items.LibraryGroupItem):
                items_list.append(item)

        return items_list

    def clear(self, *args):
        """
        Clear tree items
        """

        super(LibraryTreeWidget, self).clear(*args)
        self.clean_dirty_objects()

    def set_items(self, items):
        """
        Add given items to the tree, clearing the tree first
        :param items: list(LibraryItem)
        """

        selected_items = self.selectedItems()
        self.take_top_level_items()
        self.addTopLevelItems(items)
        self.set_items_selected(selected_items, True)

    def set_items_selected(self, items, value, scroll_to=True):
        """
        Selects the given library items
        :param items: list(LibraryItem)
        :param value: bool, Whether to select or deselect the items
        :param scroll_to: bool, Whether to scroll or not to selected items
        """

        for item in items:
            self.setItemSelected(item, value)
        if scroll_to:
            self.viewer().scroll_to_selected_item()

    def selected_item(self):
        """
        Returns the last non-hidden selected item
        :return: LibraryItem
        """

        items = self.selectedItems()
        if items:
            return items[-1]

    def settings(self):
        """
        Returns the current widget settings
        :return: dict
        """

        settings = dict()
        settings['columnSettings'] = self.column_settings()

        return settings

    def set_settings(self, settings):
        """
        Sets the current widget settings
        :param settings: dict
        :return: dict
        """

        column_settings = settings.get('columnSettings', dict())
        self.set_column_settings(column_settings)

        return settings

    def column_from_label(self, label):
        """
        Returns the column for the given label
        :param label: str
        :return: int
        """

        try:
            return self._header_labels.index(label)
        except ValueError:
            return -1

    def label_from_column(self, column):
        """
        Returns the column label for the given column
        :param column: int
        :return: str
        """

        if column is not None:
            return self.headerItem().text(column)

    def item_row(self, item):
        """
        Returns the row for the given item
        :param item: LibraryItem
        :return: int
        """

        index = self.indexFromItem(item)
        return index.row()

    def row_at(self, pos):
        """
        Returns the row for the given position
        :param pos: QPoint
        :return: int
        """

        item = self.itemAt(pos)
        return self.item_row(item)

    def take_top_level_items(self):
        """
        Returns all items from the tree widget
        :return: list(LibraryItem)
        """

        items_list = list()
        for item in self._items():
            items_list.append(self.takeTopLevelItem(1))
        items_list.append(self.takeTopLevelItem(0))

        return items_list

    def text_from_items(self, items, column, split=None, duplicates=False):
        """
        Returns all the text data for the given items and column
        :param items: list(LibraryItem)
        :param column: int or str
        :param split: str
        :param duplicates: bool
        :return: list(str)
        """

        results = list()

        for item in items:
            text = item.text(column)
            if text and split:
                results.extend(text.split(split))
            elif text:
                results.append(text)
        if not duplicates:
            results = list(set(results))

        return results

    def text_from_column(self, column, split=None, duplicates=False):
        """
        Returns all data for the given column
        :param column: int or str
        :param split: str
        :param duplicates: bool
        :return: list(str)
        """

        items_list = self.items()
        results = self.text_from_items(items_list, column, split=split, duplicates=duplicates)

        return results

    def header_labels(self):
        """
        Returns all header labels
        :return: list(str)
        """

        return self._header_labels

    def is_header_label(self, label):
        """
        Returns whether given label is a valid header label or not
        :param label: str
        :return: bool
        """

        return label in self._header_labels

    def is_sort_by_custom_order(self):
        """
        Returns True if items are currently sorted by custom order
        :return: bool
        """

        return 'Custom Order' in str(self.viewer().library().sort_by())

    def column_labels(self):
        """
        Returns all header labels for the tree widget
        :return: list(str)
        """

        return self.header_labels()

    def label_from_column(self, column):
        """
        Returns the column label for the given column
        :param column: int
        :return: str
        """

        if column is not None:
            return self.headerItem().text(column)

    def column_from_label(self, label):
        """
        Returns the column for the given label
        :param label: str
        :return: int
        """

        try:
            return self._header_labels.index(label)
        except ValueError:
            return -1

    def show_all_columns(self):
        """
        Show all available columns
        """

        for column in range(self.columnCount()):
            self.setColumnHidden(column, False)

    def hide_all_columns(self):
        """
        Hide all available columns
        """

        for column in range(1, self.columnCount()):
            self.setColumnHidden(column, True)

    def update_column_hidden(self):
        """
        Updates the hidden state for all the current columns
        """

        self.show_all_columns()
        column_labels = self._hidden_columns.keys()
        for column_label in column_labels:
            self.setColumnHidden(column_label, self._hidden_columns[column_label])

    def _items(self):
        """
        Internal function that returns a list of all items in the tree widget
        :return: list(LibraryItem)
        """

        return self.findItems('*', Qt.MatchWildcard | Qt.MatchRecursive)

    def _remove_duplicates(self, labels):
        """
        Internal function that removes dupñlicates from a list (preserving its order)
        :param labels: list(str)
        :return: list(str)
        """

        s = set()
        sadd = s.add
        return [x for x in labels if x.strip() and not (x in s or sadd(x))]

    """
    ##########################################################################################
    SETTINGS
    ##########################################################################################
    """

    def column_settings(self):
        """
        Returns the settings for each column
        :return: dict
        """

        column_settings = dict()

        for column in range(self.columnCount()):
            label = self.label_from_column(column)
            hidden = self.isColumnHidden(column)
            width = self.columnWidth(column)
            column_settings[label] = {
                'index': column,
                'hidden': hidden,
                'width': width
            }

        return column_settings

    def set_column_settings(self, settings):
        """
        Set the settings for each column
        :param settings: dict
        """

        for label in settings:
            if self.is_header_label(label):
                column = self.column_from_label(label)
                width = settings[label].get('width', 100)
                if width < 5:
                    width = 100
                self.setColumnWidth(column, width)
                hidden = settings[label].get('hidden', False)
                self.setColumnHidden(column, hidden)
            else:
                qt.logger.debug('Cannot set the column setting for header label: {}'.format(label))

    """
    ##########################################################################################
    MENUS
    ##########################################################################################
    """

    def show_header_menu(self, pos):
        """
        Creates and show the header menu at the cursor position
        :param pos: QPoint
        :return: QMenu
        """

        header = self.header()
        column = header.logicalIndexAt(pos)
        menu = self._create_header_menu(column)
        menu.addSeparator()
        sub_menu = self._create_hide_column_menu()
        menu.addMenu(sub_menu)
        menu.exec_(QCursor.pos())

    def _create_header_menu(self, column):
        """
        Internal function that creates a new header menu
        :param column, iht
        :return: QMenu
        """

        menu = QMenu(self)
        label = self.label_from_column(column)
        hide_action = menu.addAction('Hide "{}"'.format(label))
        hide_action.triggered.connect(partial(self.setColumnHidden, column, True))
        menu.addSeparator()
        resize_action = menu.addAction('Resize to Contents')
        resize_action.triggered.connect(partial(self.resizeColumnToContents, column))

        return menu

    def _create_hide_column_menu(self):
        """
        Internal function that creates the hide column menu
        :return: QMenu
        """

        menu = QMenu('Show/Hide Column', self)
        show_all_action = menu.addAction('Show All')
        show_all_action.triggered.connect(self.show_all_columns)
        hide_all_action = menu.addAction('Hide All')
        hide_all_action.triggered.connect(self.hide_all_columns)
        menu.addSeparator()
        for column in range(self.columnCount()):
            label = self.label_from_column(column)
            is_hidden = self.isColumnHidden(column)
            action = menu.addAction(label)
            action.setCheckable(True)
            action.setChecked(not is_hidden)
            action.triggered.connect(partial(self.setColumnHidden, column, not is_hidden))

        return menu

    """
    ##########################################################################################
    CLIPBOARD
    ##########################################################################################
    """

    def copy_text(self, column):
        """
        Copy the given column text to clipboard
        :param column: int or text
        """

        items = self.selectedItems()
        text = '\n'.join([item.text(column) for item in items])
        clipboard = QApplication.clipboard()
        clipboard.setText(text, QClipboard.Clipboard)

    def create_copy_text_menu(self):
        """
        Creates a menu to cpoy the selected item data to the clipboard
        :return: QMenu
        """

        menu = QMenu('Copy Text', self)
        if self.selectedItems():
            for column in range(self.columnCount()):
                label = self.label_from_column(column)
                action = menu.addAction(label)
                action.triggered.connect(partial(self.copy_text, column))
        else:
            action = menu.addAction('No items selected')
            action.setEnabled(False)

        return menu

    """
    ##########################################################################################
    CALLBACKS
    ##########################################################################################
    """

    def _on_show_header_menu(self):
        """
        Internal callback function that is called when the user right click TreeWidget
        header
        """

        print('Showing Header Menu ...')

    def _on_item_clicked(self, item):
        """
        Internal callback function that is called when an item of the tree is clicked
        """

        print('cliccked')
        item.clicked()

    def _on_item_double_clicked(self, item):
        """
        Internal callback function that is called when an item of the tree is double clicked
        """

        print('duble lcick')
        item.double_clicked()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains decorators for Qt
"""

from __future__ import print_function, division, absolute_import

from tpDcc.libs.qt.core import theme as core_theme


def theme_widget(cls):
    """
    Decorator that can be used to allow custom widget classes to retrieve theme info of its main tpDcc window
    :param cls:
    :return: cls
    """

    def theme(self):
        found_theme = None
        top_widget = self
        while top_widget.parentWidget():
            top_widget = top_widget.parentWidget()
            if hasattr(top_widget, 'theme') and callable(getattr(top_widget, 'theme')):
                found_theme = top_widget.theme()
                break

        return found_theme

    def theme_default_size(self):
        theme = self.theme()
        if not theme:
            return core_theme.Theme.DEFAULT_SIZE
        else:
            return theme.default_size

    def accent_color(self):
        theme = self.theme()
        if not theme:
            return core_theme.Theme.Colors.BLUE
        else:
            return theme.accent_color

    setattr(cls, 'theme', theme)
    setattr(cls, 'theme_default_size', theme_default_size)
    setattr(cls, 'accent_color', accent_color)

    return cls



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains awesome icons definitions
"""

from __future__ import print_function, division, absolute_import


tags = {
    "500px": u"\uf26e",
    "adjust": u"\uf042",
    "adn": u"\uf170",
    "align-center": u"\uf037",
    "align-justify": u"\uf039",
    "align-left": u"\uf036",
    "align-right": u"\uf038",
    "amazon": u"\uf270",
    "ambulance": u"\uf0f9",
    "american-sign-language-interpreting": u"\uf2a3",
    "anchor": u"\uf13d",
    "android": u"\uf17b",
    "angellist": u"\uf209",
    "angle-double-down": u"\uf103",
    "angle-double-left": u"\uf100",
    "angle-double-right": u"\uf101",
    "angle-double-up": u"\uf102",
    "angle-down": u"\uf107",
    "angle-left": u"\uf104",
    "angle-right": u"\uf105",
    "angle-up": u"\uf106",
    "apple": u"\uf179",
    "archive": u"\uf187",
    "area-chart": u"\uf1fe",
    "arrow-circle-down": u"\uf0ab",
    "arrow-circle-left": u"\uf0a8",
    "arrow-circle-o-down": u"\uf01a",
    "arrow-circle-o-left": u"\uf190",
    "arrow-circle-o-right": u"\uf18e",
    "arrow-circle-o-up": u"\uf01b",
    "arrow-circle-right": u"\uf0a9",
    "arrow-circle-up": u"\uf0aa",
    "arrow-down": u"\uf063",
    "arrow-left": u"\uf060",
    "arrow-right": u"\uf061",
    "arrow-up": u"\uf062",
    "arrows": u"\uf047",
    "arrows-alt": u"\uf0b2",
    "arrows-h": u"\uf07e",
    "arrows-v": u"\uf07d",
    "asl-interpreting (alias)": u"\uf2a3",
    "assistive-listening-systems": u"\uf2a2",
    "asterisk": u"\uf069",
    "at": u"\uf1fa",
    "audio-description": u"\uf29e",
    "automobile (alias)": u"\uf1b9",
    "backward": u"\uf04a",
    "balance-scale": u"\uf24e",
    "ban": u"\uf05e",
    "bank (alias)": u"\uf19c",
    "bar-chart": u"\uf080",
    "bar-chart-o (alias)": u"\uf080",
    "barcode": u"\uf02a",
    "bars": u"\uf0c9",
    "battery-0 (alias)": u"\uf244",
    "battery-1 (alias)": u"\uf243",
    "battery-2 (alias)": u"\uf242",
    "battery-3 (alias)": u"\uf241",
    "battery-4 (alias)": u"\uf240",
    "battery-empty": u"\uf244",
    "battery-full": u"\uf240",
    "battery-half": u"\uf242",
    "battery-quarter": u"\uf243",
    "battery-three-quarters": u"\uf241",
    "bed": u"\uf236",
    "beer": u"\uf0fc",
    "behance": u"\uf1b4",
    "behance-square": u"\uf1b5",
    "bell": u"\uf0f3",
    "bell-o": u"\uf0a2",
    "bell-slash": u"\uf1f6",
    "bell-slash-o": u"\uf1f7",
    "bicycle": u"\uf206",
    "binoculars": u"\uf1e5",
    "birthday-cake": u"\uf1fd",
    "bitbucket": u"\uf171",
    "bitbucket-square": u"\uf172",
    "bitcoin (alias)": u"\uf15a",
    "black-tie": u"\uf27e",
    "blind": u"\uf29d",
    "bluetooth": u"\uf293",
    "bluetooth-b": u"\uf294",
    "bold": u"\uf032",
    "bolt": u"\uf0e7",
    "bomb": u"\uf1e2",
    "book": u"\uf02d",
    "bookmark": u"\uf02e",
    "bookmark-o": u"\uf097",
    "braille": u"\uf2a1",
    "briefcase": u"\uf0b1",
    "btc": u"\uf15a",
    "bug": u"\uf188",
    "building": u"\uf1ad",
    "building-o": u"\uf0f7",
    "bullhorn": u"\uf0a1",
    "bullseye": u"\uf140",
    "bus": u"\uf207",
    "buysellads": u"\uf20d",
    "cab (alias)": u"\uf1ba",
    "calculator": u"\uf1ec",
    "calendar": u"\uf073",
    "calendar-check-o": u"\uf274",
    "calendar-minus-o": u"\uf272",
    "calendar-o": u"\uf133",
    "calendar-plus-o": u"\uf271",
    "calendar-times-o": u"\uf273",
    "camera": u"\uf030",
    "camera-retro": u"\uf083",
    "car": u"\uf1b9",
    "caret-down": u"\uf0d7",
    "caret-left": u"\uf0d9",
    "caret-right": u"\uf0da",
    "caret-square-o-down": u"\uf150",
    "caret-square-o-left": u"\uf191",
    "caret-square-o-right": u"\uf152",
    "caret-square-o-up": u"\uf151",
    "caret-up": u"\uf0d8",
    "cart-arrow-down": u"\uf218",
    "cart-plus": u"\uf217",
    "cc": u"\uf20a",
    "cc-amex": u"\uf1f3",
    "cc-diners-club": u"\uf24c",
    "cc-discover": u"\uf1f2",
    "cc-jcb": u"\uf24b",
    "cc-mastercard": u"\uf1f1",
    "cc-paypal": u"\uf1f4",
    "cc-stripe": u"\uf1f5",
    "cc-visa": u"\uf1f0",
    "certificate": u"\uf0a3",
    "chain (alias)": u"\uf0c1",
    "chain-broken": u"\uf127",
    "check": u"\uf00c",
    "check-circle": u"\uf058",
    "check-circle-o": u"\uf05d",
    "check-square": u"\uf14a",
    "check-square-o": u"\uf046",
    "chevron-circle-down": u"\uf13a",
    "chevron-circle-left": u"\uf137",
    "chevron-circle-right": u"\uf138",
    "chevron-circle-up": u"\uf139",
    "chevron-down": u"\uf078",
    "chevron-left": u"\uf053",
    "chevron-right": u"\uf054",
    "chevron-up": u"\uf077",
    "child": u"\uf1ae",
    "chrome": u"\uf268",
    "circle": u"\uf111",
    "circle-o": u"\uf10c",
    "circle-o-notch": u"\uf1ce",
    "circle-thin": u"\uf1db",
    "clipboard": u"\uf0ea",
    "clock-o": u"\uf017",
    "clone": u"\uf24d",
    "close (alias)": u"\uf00d",
    "cloud": u"\uf0c2",
    "cloud-download": u"\uf0ed",
    "cloud-upload": u"\uf0ee",
    "cny (alias)": u"\uf157",
    "code": u"\uf121",
    "code-fork": u"\uf126",
    "codepen": u"\uf1cb",
    "codiepie": u"\uf284",
    "coffee": u"\uf0f4",
    "cog": u"\uf013",
    "cogs": u"\uf085",
    "columns": u"\uf0db",
    "comment": u"\uf075",
    "comment-o": u"\uf0e5",
    "commenting": u"\uf27a",
    "commenting-o": u"\uf27b",
    "comments": u"\uf086",
    "comments-o": u"\uf0e6",
    "compass": u"\uf14e",
    "compress": u"\uf066",
    "connectdevelop": u"\uf20e",
    "contao": u"\uf26d",
    "copy (alias)": u"\uf0c5",
    "copyright": u"\uf1f9",
    "creative-commons": u"\uf25e",
    "credit-card": u"\uf09d",
    "credit-card-alt": u"\uf283",
    "crop": u"\uf125",
    "crosshairs": u"\uf05b",
    "css3": u"\uf13c",
    "cube": u"\uf1b2",
    "cubes": u"\uf1b3",
    "cut (alias)": u"\uf0c4",
    "cutlery": u"\uf0f5",
    "dashboard (alias)": u"\uf0e4",
    "dashcube": u"\uf210",
    "database": u"\uf1c0",
    "deaf": u"\uf2a4",
    "deafness (alias)": u"\uf2a4",
    "dedent (alias)": u"\uf03b",
    "delicious": u"\uf1a5",
    "desktop": u"\uf108",
    "deviantart": u"\uf1bd",
    "diamond": u"\uf219",
    "digg": u"\uf1a6",
    "dollar (alias)": u"\uf155",
    "dot-circle-o": u"\uf192",
    "download": u"\uf019",
    "dribbble": u"\uf17d",
    "dropbox": u"\uf16b",
    "drupal": u"\uf1a9",
    "edge": u"\uf282",
    "edit (alias)": u"\uf044",
    "eject": u"\uf052",
    "ellipsis-h": u"\uf141",
    "ellipsis-v": u"\uf142",
    "empire": u"\uf1d1",
    "envelope": u"\uf0e0",
    "envelope-o": u"\uf003",
    "envelope-square": u"\uf199",
    "envira": u"\uf299",
    "eraser": u"\uf12d",
    "eur": u"\uf153",
    "euro (alias)": u"\uf153",
    "exchange": u"\uf0ec",
    "exclamation": u"\uf12a",
    "exclamation-circle": u"\uf06a",
    "exclamation-triangle": u"\uf071",
    "expand": u"\uf065",
    "expeditedssl": u"\uf23e",
    "external-link": u"\uf08e",
    "external-link-square": u"\uf14c",
    "eye": u"\uf06e",
    "eye-slash": u"\uf070",
    "eyedropper": u"\uf1fb",
    "fa (alias)": u"\uf2b4",
    "facebook": u"\uf09a",
    "facebook-f (alias)": u"\uf09a",
    "facebook-official": u"\uf230",
    "facebook-square": u"\uf082",
    "fast-backward": u"\uf049",
    "fast-forward": u"\uf050",
    "fax": u"\uf1ac",
    "feed (alias)": u"\uf09e",
    "female": u"\uf182",
    "fighter-jet": u"\uf0fb",
    "file": u"\uf15b",
    "file-archive-o": u"\uf1c6",
    "file-audio-o": u"\uf1c7",
    "file-code-o": u"\uf1c9",
    "file-excel-o": u"\uf1c3",
    "file-image-o": u"\uf1c5",
    "file-movie-o (alias)": u"\uf1c8",
    "file-o": u"\uf016",
    "file-pdf-o": u"\uf1c1",
    "file-photo-o (alias)": u"\uf1c5",
    "file-picture-o (alias)": u"\uf1c5",
    "file-powerpoint-o": u"\uf1c4",
    "file-sound-o (alias)": u"\uf1c7",
    "file-text": u"\uf15c",
    "file-text-o": u"\uf0f6",
    "file-video-o": u"\uf1c8",
    "file-word-o": u"\uf1c2",
    "file-zip-o (alias)": u"\uf1c6",
    "files-o": u"\uf0c5",
    "film": u"\uf008",
    "filter": u"\uf0b0",
    "fire": u"\uf06d",
    "fire-extinguisher": u"\uf134",
    "firefox": u"\uf269",
    "first-order": u"\uf2b0",
    "flag": u"\uf024",
    "flag-checkered": u"\uf11e",
    "flag-o": u"\uf11d",
    "flash (alias)": u"\uf0e7",
    "flask": u"\uf0c3",
    "flickr": u"\uf16e",
    "floppy-o": u"\uf0c7",
    "folder": u"\uf07b",
    "folder-o": u"\uf114",
    "folder-open": u"\uf07c",
    "folder-open-o": u"\uf115",
    "font": u"\uf031",
    "font-awesome": u"\uf2b4",
    "fonticons": u"\uf280",
    "fort-awesome": u"\uf286",
    "forumbee": u"\uf211",
    "forward": u"\uf04e",
    "foursquare": u"\uf180",
    "frown-o": u"\uf119",
    "futbol-o": u"\uf1e3",
    "gamepad": u"\uf11b",
    "gavel": u"\uf0e3",
    "gbp": u"\uf154",
    "ge (alias)": u"\uf1d1",
    "gear (alias)": u"\uf013",
    "gears (alias)": u"\uf085",
    "genderless": u"\uf22d",
    "get-pocket": u"\uf265",
    "gg": u"\uf260",
    "gg-circle": u"\uf261",
    "gift": u"\uf06b",
    "git": u"\uf1d3",
    "git-square": u"\uf1d2",
    "github": u"\uf09b",
    "github-alt": u"\uf113",
    "github-square": u"\uf092",
    "gitlab": u"\uf296",
    "gittip (alias)": u"\uf184",
    "glass": u"\uf000",
    "glide": u"\uf2a5",
    "glide-g": u"\uf2a6",
    "globe": u"\uf0ac",
    "google": u"\uf1a0",
    "google-plus": u"\uf0d5",
    "google-plus-circle (alias)": u"\uf2b3",
    "google-plus-official": u"\uf2b3",
    "google-plus-square": u"\uf0d4",
    "google-wallet": u"\uf1ee",
    "graduation-cap": u"\uf19d",
    "gratipay": u"\uf184",
    "group (alias)": u"\uf0c0",
    "h-square": u"\uf0fd",
    "hacker-news": u"\uf1d4",
    "hand-grab-o (alias)": u"\uf255",
    "hand-lizard-o": u"\uf258",
    "hand-o-down": u"\uf0a7",
    "hand-o-left": u"\uf0a5",
    "hand-o-right": u"\uf0a4",
    "hand-o-up": u"\uf0a6",
    "hand-paper-o": u"\uf256",
    "hand-peace-o": u"\uf25b",
    "hand-pointer-o": u"\uf25a",
    "hand-rock-o": u"\uf255",
    "hand-scissors-o": u"\uf257",
    "hand-spock-o": u"\uf259",
    "hand-stop-o (alias)": u"\uf256",
    "hard-of-hearing (alias)": u"\uf2a4",
    "hashtag": u"\uf292",
    "hdd-o": u"\uf0a0",
    "header": u"\uf1dc",
    "headphones": u"\uf025",
    "heart": u"\uf004",
    "heart-o": u"\uf08a",
    "heartbeat": u"\uf21e",
    "history": u"\uf1da",
    "home": u"\uf015",
    "hospital-o": u"\uf0f8",
    "hotel (alias)": u"\uf236",
    "hourglass": u"\uf254",
    "hourglass-1 (alias)": u"\uf251",
    "hourglass-2 (alias)": u"\uf252",
    "hourglass-3 (alias)": u"\uf253",
    "hourglass-end": u"\uf253",
    "hourglass-half": u"\uf252",
    "hourglass-o": u"\uf250",
    "hourglass-start": u"\uf251",
    "houzz": u"\uf27c",
    "html5": u"\uf13b",
    "i-cursor": u"\uf246",
    "ils": u"\uf20b",
    "image (alias)": u"\uf03e",
    "inbox": u"\uf01c",
    "indent": u"\uf03c",
    "industry": u"\uf275",
    "info": u"\uf129",
    "info-circle": u"\uf05a",
    "inr": u"\uf156",
    "instagram": u"\uf16d",
    "institution (alias)": u"\uf19c",
    "internet-explorer": u"\uf26b",
    "intersex (alias)": u"\uf224",
    "ioxhost": u"\uf208",
    "italic": u"\uf033",
    "joomla": u"\uf1aa",
    "jpy": u"\uf157",
    "jsfiddle": u"\uf1cc",
    "key": u"\uf084",
    "keyboard-o": u"\uf11c",
    "krw": u"\uf159",
    "language": u"\uf1ab",
    "laptop": u"\uf109",
    "lastfm": u"\uf202",
    "lastfm-square": u"\uf203",
    "leaf": u"\uf06c",
    "leanpub": u"\uf212",
    "legal (alias)": u"\uf0e3",
    "lemon-o": u"\uf094",
    "level-down": u"\uf149",
    "level-up": u"\uf148",
    "life-bouy (alias)": u"\uf1cd",
    "life-buoy (alias)": u"\uf1cd",
    "life-ring": u"\uf1cd",
    "life-saver (alias)": u"\uf1cd",
    "lightbulb-o": u"\uf0eb",
    "line-chart": u"\uf201",
    "link": u"\uf0c1",
    "linkedin": u"\uf0e1",
    "linkedin-square": u"\uf08c",
    "linux": u"\uf17c",
    "list": u"\uf03a",
    "list-alt": u"\uf022",
    "list-ol": u"\uf0cb",
    "list-ul": u"\uf0ca",
    "location-arrow": u"\uf124",
    "lock": u"\uf023",
    "long-arrow-down": u"\uf175",
    "long-arrow-left": u"\uf177",
    "long-arrow-right": u"\uf178",
    "long-arrow-up": u"\uf176",
    "low-vision": u"\uf2a8",
    "magic": u"\uf0d0",
    "magnet": u"\uf076",
    "mail-forward (alias)": u"\uf064",
    "mail-reply (alias)": u"\uf112",
    "mail-reply-all (alias)": u"\uf122",
    "male": u"\uf183",
    "map": u"\uf279",
    "map-marker": u"\uf041",
    "map-o": u"\uf278",
    "map-pin": u"\uf276",
    "map-signs": u"\uf277",
    "mars": u"\uf222",
    "mars-double": u"\uf227",
    "mars-stroke": u"\uf229",
    "mars-stroke-h": u"\uf22b",
    "mars-stroke-v": u"\uf22a",
    "maxcdn": u"\uf136",
    "meanpath": u"\uf20c",
    "medium": u"\uf23a",
    "medkit": u"\uf0fa",
    "meh-o": u"\uf11a",
    "mercury": u"\uf223",
    "microphone": u"\uf130",
    "microphone-slash": u"\uf131",
    "minus": u"\uf068",
    "minus-circle": u"\uf056",
    "minus-square": u"\uf146",
    "minus-square-o": u"\uf147",
    "mixcloud": u"\uf289",
    "mobile": u"\uf10b",
    "mobile-phone (alias)": u"\uf10b",
    "modx": u"\uf285",
    "money": u"\uf0d6",
    "moon-o": u"\uf186",
    "mortar-board (alias)": u"\uf19d",
    "motorcycle": u"\uf21c",
    "mouse-pointer": u"\uf245",
    "music": u"\uf001",
    "navicon (alias)": u"\uf0c9",
    "neuter": u"\uf22c",
    "newspaper-o": u"\uf1ea",
    "object-group": u"\uf247",
    "object-ungroup": u"\uf248",
    "odnoklassniki": u"\uf263",
    "odnoklassniki-square": u"\uf264",
    "opencart": u"\uf23d",
    "openid": u"\uf19b",
    "opera": u"\uf26a",
    "optin-monster": u"\uf23c",
    "outdent": u"\uf03b",
    "pagelines": u"\uf18c",
    "paint-brush": u"\uf1fc",
    "paper-plane": u"\uf1d8",
    "paper-plane-o": u"\uf1d9",
    "paperclip": u"\uf0c6",
    "paragraph": u"\uf1dd",
    "paste (alias)": u"\uf0ea",
    "pause": u"\uf04c",
    "pause-circle": u"\uf28b",
    "pause-circle-o": u"\uf28c",
    "paw": u"\uf1b0",
    "paypal": u"\uf1ed",
    "pencil": u"\uf040",
    "pencil-square": u"\uf14b",
    "pencil-square-o": u"\uf044",
    "percent": u"\uf295",
    "phone": u"\uf095",
    "phone-square": u"\uf098",
    "photo (alias)": u"\uf03e",
    "picture-o": u"\uf03e",
    "pie-chart": u"\uf200",
    "pied-piper": u"\uf2ae",
    "pied-piper-alt": u"\uf1a8",
    "pied-piper-pp": u"\uf1a7",
    "pinterest": u"\uf0d2",
    "pinterest-p": u"\uf231",
    "pinterest-square": u"\uf0d3",
    "plane": u"\uf072",
    "play": u"\uf04b",
    "play-circle": u"\uf144",
    "play-circle-o": u"\uf01d",
    "plug": u"\uf1e6",
    "plus": u"\uf067",
    "plus-circle": u"\uf055",
    "plus-square": u"\uf0fe",
    "plus-square-o": u"\uf196",
    "power-off": u"\uf011",
    "print": u"\uf02f",
    "product-hunt": u"\uf288",
    "puzzle-piece": u"\uf12e",
    "qq": u"\uf1d6",
    "qrcode": u"\uf029",
    "question": u"\uf128",
    "question-circle": u"\uf059",
    "question-circle-o": u"\uf29c",
    "quote-left": u"\uf10d",
    "quote-right": u"\uf10e",
    "ra (alias)": u"\uf1d0",
    "random": u"\uf074",
    "rebel": u"\uf1d0",
    "recycle": u"\uf1b8",
    "reddit": u"\uf1a1",
    "reddit-alien": u"\uf281",
    "reddit-square": u"\uf1a2",
    "refresh": u"\uf021",
    "registered": u"\uf25d",
    "remove (alias)": u"\uf00d",
    "renren": u"\uf18b",
    "reorder (alias)": u"\uf0c9",
    "repeat": u"\uf01e",
    "reply": u"\uf112",
    "reply-all": u"\uf122",
    "resistance (alias)": u"\uf1d0",
    "retweet": u"\uf079",
    "rmb (alias)": u"\uf157",
    "road": u"\uf018",
    "rocket": u"\uf135",
    "rotate-left (alias)": u"\uf0e2",
    "rotate-right (alias)": u"\uf01e",
    "rouble (alias)": u"\uf158",
    "rss": u"\uf09e",
    "rss-square": u"\uf143",
    "rub": u"\uf158",
    "ruble (alias)": u"\uf158",
    "rupee (alias)": u"\uf156",
    "safari": u"\uf267",
    "save (alias)": u"\uf0c7",
    "scissors": u"\uf0c4",
    "scribd": u"\uf28a",
    "search": u"\uf002",
    "search-minus": u"\uf010",
    "search-plus": u"\uf00e",
    "sellsy": u"\uf213",
    "send (alias)": u"\uf1d8",
    "send-o (alias)": u"\uf1d9",
    "server": u"\uf233",
    "share": u"\uf064",
    "share-alt": u"\uf1e0",
    "share-alt-square": u"\uf1e1",
    "share-square": u"\uf14d",
    "share-square-o": u"\uf045",
    "shekel (alias)": u"\uf20b",
    "sheqel (alias)": u"\uf20b",
    "shield": u"\uf132",
    "ship": u"\uf21a",
    "shirtsinbulk": u"\uf214",
    "shopping-bag": u"\uf290",
    "shopping-basket": u"\uf291",
    "shopping-cart": u"\uf07a",
    "sign-in": u"\uf090",
    "sign-language": u"\uf2a7",
    "sign-out": u"\uf08b",
    "signal": u"\uf012",
    "signing (alias)": u"\uf2a7",
    "simplybuilt": u"\uf215",
    "sitemap": u"\uf0e8",
    "skyatlas": u"\uf216",
    "skype": u"\uf17e",
    "slack": u"\uf198",
    "sliders": u"\uf1de",
    "slideshare": u"\uf1e7",
    "smile-o": u"\uf118",
    "snapchat": u"\uf2ab",
    "snapchat-ghost": u"\uf2ac",
    "snapchat-square": u"\uf2ad",
    "soccer-ball-o (alias)": u"\uf1e3",
    "sort": u"\uf0dc",
    "sort-alpha-asc": u"\uf15d",
    "sort-alpha-desc": u"\uf15e",
    "sort-amount-asc": u"\uf160",
    "sort-amount-desc": u"\uf161",
    "sort-asc": u"\uf0de",
    "sort-desc": u"\uf0dd",
    "sort-down (alias)": u"\uf0dd",
    "sort-numeric-asc": u"\uf162",
    "sort-numeric-desc": u"\uf163",
    "sort-up (alias)": u"\uf0de",
    "soundcloud": u"\uf1be",
    "space-shuttle": u"\uf197",
    "spinner": u"\uf110",
    "spoon": u"\uf1b1",
    "spotify": u"\uf1bc",
    "square": u"\uf0c8",
    "square-o": u"\uf096",
    "stack-exchange": u"\uf18d",
    "stack-overflow": u"\uf16c",
    "star": u"\uf005",
    "star-half": u"\uf089",
    "star-half-empty (alias)": u"\uf123",
    "star-half-full (alias)": u"\uf123",
    "star-half-o": u"\uf123",
    "star-o": u"\uf006",
    "steam": u"\uf1b6",
    "steam-square": u"\uf1b7",
    "step-backward": u"\uf048",
    "step-forward": u"\uf051",
    "stethoscope": u"\uf0f1",
    "sticky-note": u"\uf249",
    "sticky-note-o": u"\uf24a",
    "stop": u"\uf04d",
    "stop-circle": u"\uf28d",
    "stop-circle-o": u"\uf28e",
    "street-view": u"\uf21d",
    "strikethrough": u"\uf0cc",
    "stumbleupon": u"\uf1a4",
    "stumbleupon-circle": u"\uf1a3",
    "subscript": u"\uf12c",
    "subway": u"\uf239",
    "suitcase": u"\uf0f2",
    "sun-o": u"\uf185",
    "superscript": u"\uf12b",
    "support (alias)": u"\uf1cd",
    "table": u"\uf0ce",
    "tablet": u"\uf10a",
    "tachometer": u"\uf0e4",
    "tag": u"\uf02b",
    "tags": u"\uf02c",
    "tasks": u"\uf0ae",
    "taxi": u"\uf1ba",
    "television": u"\uf26c",
    "tencent-weibo": u"\uf1d5",
    "terminal": u"\uf120",
    "text-height": u"\uf034",
    "text-width": u"\uf035",
    "th": u"\uf00a",
    "th-large": u"\uf009",
    "th-list": u"\uf00b",
    "themeisle": u"\uf2b2",
    "thumb-tack": u"\uf08d",
    "thumbs-down": u"\uf165",
    "thumbs-o-down": u"\uf088",
    "thumbs-o-up": u"\uf087",
    "thumbs-up": u"\uf164",
    "ticket": u"\uf145",
    "times": u"\uf00d",
    "times-circle": u"\uf057",
    "times-circle-o": u"\uf05c",
    "tint": u"\uf043",
    "toggle-down (alias)": u"\uf150",
    "toggle-left (alias)": u"\uf191",
    "toggle-off": u"\uf204",
    "toggle-on": u"\uf205",
    "toggle-right (alias)": u"\uf152",
    "toggle-up (alias)": u"\uf151",
    "trademark": u"\uf25c",
    "train": u"\uf238",
    "transgender": u"\uf224",
    "transgender-alt": u"\uf225",
    "trash": u"\uf1f8",
    "trash-o": u"\uf014",
    "tree": u"\uf1bb",
    "trello": u"\uf181",
    "tripadvisor": u"\uf262",
    "trophy": u"\uf091",
    "truck": u"\uf0d1",
    "try": u"\uf195",
    "tty": u"\uf1e4",
    "tumblr": u"\uf173",
    "tumblr-square": u"\uf174",
    "turkish-lira (alias)": u"\uf195",
    "tv (alias)": u"\uf26c",
    "twitch": u"\uf1e8",
    "twitter": u"\uf099",
    "twitter-square": u"\uf081",
    "umbrella": u"\uf0e9",
    "underline": u"\uf0cd",
    "undo": u"\uf0e2",
    "universal-access": u"\uf29a",
    "university": u"\uf19c",
    "unlink (alias)": u"\uf127",
    "unlock": u"\uf09c",
    "unlock-alt": u"\uf13e",
    "unsorted (alias)": u"\uf0dc",
    "upload": u"\uf093",
    "usb": u"\uf287",
    "usd": u"\uf155",
    "user": u"\uf007",
    "user-md": u"\uf0f0",
    "user-plus": u"\uf234",
    "user-secret": u"\uf21b",
    "user-times": u"\uf235",
    "users": u"\uf0c0",
    "venus": u"\uf221",
    "venus-double": u"\uf226",
    "venus-mars": u"\uf228",
    "viacoin": u"\uf237",
    "viadeo": u"\uf2a9",
    "viadeo-square": u"\uf2aa",
    "video-camera": u"\uf03d",
    "vimeo": u"\uf27d",
    "vimeo-square": u"\uf194",
    "vine": u"\uf1ca",
    "vk": u"\uf189",
    "volume-control-phone": u"\uf2a0",
    "volume-down": u"\uf027",
    "volume-off": u"\uf026",
    "volume-up": u"\uf028",
    "warning (alias)": u"\uf071",
    "wechat (alias)": u"\uf1d7",
    "weibo": u"\uf18a",
    "weixin": u"\uf1d7",
    "whatsapp": u"\uf232",
    "wheelchair": u"\uf193",
    "wheelchair-alt": u"\uf29b",
    "wifi": u"\uf1eb",
    "wikipedia-w": u"\uf266",
    "windows": u"\uf17a",
    "won (alias)": u"\uf159",
    "wordpress": u"\uf19a",
    "wpbeginner": u"\uf297",
    "wpforms": u"\uf298",
    "wrench": u"\uf0ad",
    "xing": u"\uf168",
    "xing-square": u"\uf169",
    "y-combinator": u"\uf23b",
    "y-combinator-square (alias)": u"\uf1d4",
    "yahoo": u"\uf19e",
    "yc (alias)": u"\uf23b",
    "yc-square (alias)": u"\uf1d4",
    "yelp": u"\uf1e9",
    "yen (alias)": u"\uf157",
    "yoast": u"\uf2b1",
    "youtube": u"\uf167",
    "youtube-play": u"\uf16a",
    "youtube-square": u"\uf166",
}



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt tab widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *

import tpDcc as tp
from tpDcc.libs.qt.core import qtutils, mixin
from tpDcc.libs.qt.core import base, window, theme
from tpDcc.libs.qt.widgets import tabbars, buttons, group, dividers, stack


class BaseTabWidget(QTabWidget, object):
    """
    Base class for tab widgets
    """

    deletedTab = Signal(object)

    def __init__(self, parent=None):
        super(BaseTabWidget, self).__init__(parent=parent)

        self._set_custom_tab_bar()
        self.tabCloseRequested.connect(self.removeTab)

        self._show_close_message = False
        self._show_text = ''
        self._show_title = ''
        self._show_fn = None
        self._show_can_cancel = False

    def removeTab(self, index):
        """
        Function that is called when a tab is removed
        :param index: int, index of the removed tab
        """

        if self._show_close_message:
            can_close = qtutils.get_permission(
                message='Save graph before closing it?',
                cancel=self._show_can_cancel, title='Closing Graph', parent=self)
            if can_close is None:
                return
            elif can_close:
                if self._show_fn:
                    self._show_fn(index, self.widget(index))

        super(BaseTabWidget, self).removeTab(index)
        self.deletedTab.emit(index)

    def set_close_message(self, message_text, message_title, cancel=False, accept_fn_call=None):
        self._show_close_message = True
        self._show_text = message_text
        self._show_title = message_title
        self._show_can_cancel = cancel
        self._show_fn = accept_fn_call

    def get_unique_tab_name(self, name, curr_index=0):
        """
        Returns a unique tab name for the current tab widget
        :param name: str, base name to get a unique name
        :param curr_index: int, starting index to start searching for unique names
        :return:
        """

        curr_index = curr_index
        new_name = name
        index = 0
        if curr_index > 0:
            new_name = name + '_' + str(curr_index)
        for i in range(self.tabBar().count()):
            curr_tab_name = self.tabBar().tabText(i)
            if curr_tab_name == new_name:
                index = index + 1
        if index > 0:
            return self.get_unique_tab_name(name, curr_index=curr_index + index)
        return new_name

    def update_tab_bar_signals(self):
        """
        Override functionality if necessary
        Updates teh tab bar signals
        """

        pass

    def _set_custom_tab_bar(self):
        """
        Sets a custom tab bar with editable functionality for the widget
        """

        pass


class BaseEditableTabWidget(BaseTabWidget, object):
    """
    Tab widget with basic editable functionality
    """

    def __init__(self, parent=None):
        super(BaseEditableTabWidget, self).__init__(parent=parent)

        self.setMovable(True)
        self.setTabsClosable(True)
        self._edit_mode = True

    def set_edit_mode(self, value):
        if self._edit_mode != value:
            self._edit_mode = value
            self.setMovable(value)
            self.setTabsClosable(value)

    def get_edit_mode(self):
        return self._edit_mode

    edit_mode = property(get_edit_mode, set_edit_mode)

    def mousePressEvent(self, event):
        pos = event.pos()
        rect = self._add_tab_btn_rect()
        if rect.contains(pos):
            pass
        # super(BaseEditableTabWidget, self).mousePressEvent(event)

    def _add_tab_btn_rect(self):
        """
        Returns the visual rect of the current selected tab
        :return: QRect
        """

        return self.tabBar().add_tab_btn.rect()

    def _set_custom_tab_bar(self):
        """
        Sets a custom tab bar with editable functionality for the widget
        """

        self.setTabBar(tabbars.EditableTabBar(self))
        self.update_tab_bar_signals()


class TearOffTabWidget(BaseTabWidget, object):
    """
    Tab widget with basic tear off functionality
    """

    class TabWidgetStorage(object):
        def __init__(self, index, widget, title, visible=True):
            self.index = index
            self.widget = widget
            self.title = title
            self.visible = visible
            self.detached = None

        def set_detached(self, detached):
            self.detached = detached

        def __repr__(self):
            return 'index %d, widget %r, title %s, visible %r, detached %r' % (
                self.index, self.widget, self.title, self.visible, self.detached)

    tabAdded = Signal()
    tabDetached = Signal(int, QPoint)
    tabBarRenamed = Signal(str, str)
    requestRemove = Signal(int)
    tabChanged = Signal(int)
    undoOpen = Signal()
    undoClose = Signal()

    # def __init__(self, item, window, menu_window, parent=None):
    def __init__(self, parent=None):
        super(TearOffTabWidget, self).__init__(parent=parent)

        self._previous_tab_id = 0
        self.tab_idx = dict()

        self.setMovable(False)
        self.setTabsClosable(True)
        self.setMouseTracking(True)
        self.setObjectName('BaseTabWidget')

        self._set_palette_color()

        # don't draw a frame around the tab contents
        self.setStyleSheet('QTabWidget:tab:disabled{width:0;height:0;margin:0;padding:0;border:none}')
        self.setDocumentMode(True)

    def update_tab_bar_signals(self):
        """
        Updates the signals attached to the tab bar
        """

        tab_bar = self.tabBar()
        if tab_bar is None:
            return
        self.tabBar().tabDetached.connect(self.tabDetached.emit)
        self.tabBar().tabDetached.connect(self._on_detach_tab)
        self.tabBar().tabMoved.connect(self._on_move_tab)
        self.currentChanged.connect(self._on_tab_changed_tab)

    def addTab(self, *args, **kwargs):
        added_tab_index = super(TearOffTabWidget, self).addTab(*args, **kwargs)
        self.tab_inserted(added_tab_index)

    def removeTab(self, index):
        super(TearOffTabWidget, self).removeTab(index)
        self.tab_idx.pop(index)

    def tab_inserted(self, index):
        w = self.widget(index)
        for i in self.tab_idx.values():
            if i.widget == w:
                return
        self.tab_idx[index] = self.TabWidgetStorage(index, self.widget(index), self.tabText(index))

    def set_widget_visible(self, widget, visible):
        w = self._find_first_window(widget)
        for i in self.tab_idx.values():
            if i.widget == w:
                if visible:
                    if not i.visible:
                        new_index = -1
                        for j in self.tab_idx.values():
                            if j.visible and j.index > i.index:
                                c_idx = self._tab_widget_index(j.widget)
                                if c_idx < i.index and c_idx != -1:
                                    new_index = c_idx
                                else:
                                    new_index = i.index
                                break
                        self.insertTab(new_index, i.widget, i.title)
                else:
                    i.visible = False
                    index = self._tab_widget_index(i.widget)
                    self.removeTab(index)

    def add_panel(self, widget, label):

        # TODO: We should not storing the detach config here
        # sg_group = window.DetachedWindow.SettingGroup(label)
        # with sg_group as settings:
        #     detached = settings.value('detached', False)
        detached = False

        index = len(self.tab_idx)
        self.tab_idx[index] = self.TabWidgetStorage(index, widget, label)
        if not detached:
            index = self.addTab(widget, label)
            for i in self.tab_idx.values():
                if i.widget == widget:
                    i.set_detached(None)
        else:
            detach_window = window.DetachedWindow(label, self.parentWidget())
            detach_window.tab_idx = index
            detach_window.setAttribute(Qt.WA_DeleteOnClose, True)
            self.tab_idx[index].set_detached(detach_window)
            detach_window.windowClosed.connect(self._on_attach_tab)

            panel = self._get_widget(widget)
            panel.widgetVisible.disconnect(self.set_widget_visible)
            panel.widgetVisible.connect(detach_window.set_widget_visible)
            widget.setParent(detach_window)

            detach_window.set_widget(widget)
            detach_window.destroyed.connect(detach_window.deleteLater)
            # detach_window.restoreGeometry(settings.value('geometry', ''))
            detach_window.show()

    def clear(self):
        """
        Delete all the tabs added to this tab widget
        """

        all_tabs = [self.widget(i) for i in range(self.count())]
        self.blockSignals(True)
        for tab in all_tabs:
            tab.deleteLater()
        super(TearOffTabWidget, self).clear()
        self.blockSignals(False)

    def dragEnterEvent(self, event):
        """
        Override drag enter behaviour
        """

        event.accept()
        super(TearOffTabWidget, self).dragEnterEvent(event)

    def dragMoveEvent(self, event):
        """
        Override drag move behaviour
        """

        event.accept()
        super(TearOffTabWidget, self).dragMoveEvent(event)
    # endregion

    # region Private Functions
    def _set_palette_color(self):
        """
        Sets the initial for the different elements of the tab widget
        Override to create custom color styles for this widget
        """

        palette = self.palette()
        self.border_color = QColor(Qt.black)
        self.button_color = palette.color(QPalette.Button)
        self.fill_color = palette.color(QPalette.Window)
        self.highlight_color = palette.color(QPalette.Highlight)

    def _set_custom_tab_bar(self):
        """
        Sets a custom tab bar with tear off functionality for the widget
        Override to add a custom tab bar for this widget
        """

        tab_bar = tabbars.TearOffTabBar()
        self.setTabBar(tab_bar)
        self.update_tab_bar_signals()

    def _find_first_window(self, w):
        widget = w
        while True:
            parent = widget.parent()
            if not parent:
                break
            widget = parent
            if isinstance(widget, QMainWindow):
                break

        return widget

    def _tab_widget_index(self, widget):
        for i in range(self.tabBar().count()):
            if self.widget(i) == widget:
                return i

        return -1

    def _get_widget(self, widget):
        panel = widget
        if isinstance(widget, QMainWindow):
            panel = widget.centralWidget()
            if panel.layout():
                panel = panel.layout().itemAt(0).widget()
        elif isinstance(widget, window.DetachedWindow.DetachPanel):
            if panel.layout():
                panel = panel.layout().itemAt(0).widget()

        return panel

    def _on_attach_tab(self, detach_window):
        detach_window.windowClosed.disconnect(self._on_attach_tab)
        detach_window.save_settings(False)
        panel = detach_window.centralWidget()
        tear_off_widget = self._get_widget(panel)
        try:
            panel.widgetVisible.disconnect(detach_window.set_widget_visible)
        except Exception:
            pass
        tear_off_widget.setParent(self)
        panel.widgetVisible.connect(self.set_widget_visible)

        new_index = -1

        for i in range(self.tabBar().count()):
            w = self.widget(i)
            for j in self.tab_idx.values():
                if j.widget == w and j.index > detach_window.tab_idx:
                    new_index = i
                    break
            else:
                continue
            break

        if new_index == -1:
            new_index = self.tabBar().count()

        new_index = self.insertTab(new_index, tear_off_widget, detach_window.windowTitle())

        if new_index != -1:
            self.setCurrentIndex(new_index)

        self.tab_idx[new_index].detached = None

    def _on_detach_tab(self, index, point):
        detach_window = window.DetachedWindow(self.tabText(index), self.parentWidget())
        w = self.widget(index)
        for i in self.tab_idx.values():
            if i.widget == w:
                detach_window.tab_idx = self.tab_idx[i.index].index
                self.tab_idx[i.index].detached = detach_window
                break

        detach_window.windowClosed.connect(self._on_attach_tab)

        tear_off_widget = self.widget(index)
        # panel = detach_window.centralWidget()
        # panel.widgetVisible.disconnect(self.set_widget_visible)
        # panel.widgetVisible.connect(detach_window.set_widget_visible)
        tear_off_widget.setParent(detach_window)

        if self.count() < 0:
            self.setCurrentIndex(0)

        detach_window.set_widget(tear_off_widget)
        detach_window.resize(tear_off_widget.size())
        detach_window.move(point)
        detach_window.show()

    def _on_move_tab(self, from_index, to_index):
        w = self.widget(from_index)
        text = self.tabText(from_index)
        self.removeTab(from_index)
        self.insertTab(to_index, w, text)
        self.setCurrentIndex(to_index)

    def _on_tab_changed_tab(self, index):
        self._previous_tab_id = index


class EditableTearOffTabWidget(TearOffTabWidget, object):
    """
    Custom tab widget with editable functionality
    """

    tab_size_changed = Signal(QSize)
    tab_label_renamed = Signal(str, str)
    tab_removed_text = Signal(str)
    window_modified = Signal()
    add_item_on = Signal(QPointF, QColor, int)
    add_item_with_set = Signal(QPointF, QColor, str, int)
    add_previewed_item = Signal(QPointF, QColor, str, int)
    copy_items = Signal(list)
    selected_items_changed = Signal(list)

    def __init__(self, parent=None):
        super(EditableTearOffTabWidget, self).__init__(parent=parent)

        self._is_editable = True
        self._has_highlight = False
        self.setAcceptDrops(True)
        self.setObjectName('EditableTabWidget')

    def set_is_editable(self, value):
        if self._is_editable != value:
            self._is_editable = value
            self.setMovable(value)
            self.update()

    def get_is_editable(self):
        return self._is_editable

    def set_has_highlight(self, value):
        if self._has_highlight != value:
            self._has_highlight = value
            self.update()

    def get_has_highlight(self):
        return self._has_highlight

    is_editable = property(get_is_editable, set_is_editable)
    has_highlight = property(get_has_highlight, set_has_highlight)

    def toggle_editable(self):
        """
        Toggles editable property of the tab widget
        """

        self.is_editable = not self.is_editable

    def delete_tab(self, index):
        """
        Deletes a tab with given index
        :param index: index of the tab to delete
        """

        if index < 0:
            return
        widget = self.widget(index)
        tab_text = self.tabText(index)
        super(EditableTearOffTabWidget, self).removeTab(index)
        widget.deleteLater()
        self.window_modified.emit()
        self.tab_removed_text.emit(tab_text)

    def button_rect(self):
        r = self.tabBar().tabRect(self.count() - 1)

        if tp.Dcc.get_name() == tp.Dccs.Maya and tp.Dcc.get_version() > 2015:
            rect = QRect(2, 0, 30, 31)
        else:
            rect = QRect(6, 3, 26, 17)
        if r.isValid():
            if tp.Dcc.get_name() == tp.Dccs.Maya and tp.Dcc.get_version() > 2015:
                rect.moveBottomLeft(r.bottomRight() + QPoint(1, 1))
            else:
                rect.moveBottomLeft(r.bottomRight() + QPoint(3, -1))
        return rect

    def update_tab_bar_signals(self):
        """
        Updates the signals attached to the tab bar
        """

        super(EditableTearOffTabWidget, self).update_tab_bar_signals()
        tab_bar = self.tabBar()
        if tab_bar is None:
            return
        tab_bar.tabMoved.connect(lambda: self.window_modified.emit())
        tab_bar.tab_label_renamed.connect(lambda: self.window_modified.emit())
        tab_bar.tab_label_renamed.connect(self.tab_label_renamed.emit)
        tab_bar.request_remove.connect(self.delete_tab)

    def _set_custom_tab_bar(self):
        """
        Sets a custom tab bar with editable and tear off functionality for the widget
        """

        tabBar = tabbars.EditableTearOffTabBar(self)
        self.setTabBar(tabBar)
        self.update_tab_bar_signals()
        self.setTabsClosable(True)


class MenuTabBlockButton(buttons.BaseToolButton, object):
    def __init__(self, parent=None):
        super(MenuTabBlockButton, self).__init__(parent=parent)

        self.setCheckable(True)


@mixin.theme_mixin
class MenuTabBlockButtonGroup(group.BaseButtonGroup, object):
    checkedChanged = Signal(int)

    def __init_(self, parent=None):
        super(MenuTabBlockButtonGroup, self).__init__(parent=parent)

        self.set_spacing(1)
        self._button_group.setExclusive(True)
        self._button_group.buttonClicked[int].connect(self.checkedChanged)

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_checked(self):
        """
        Returns current checked button's ID
        :return:  int
        """

        return self._button_group.checkedId()

    def _set_checked(self, value):
        """
        Sets current checked button's ID
        :param value: int
        """

        btn = self._button_group.button(value)
        btn.setChecked(True)
        self.checkedChanged.emit(value)

    checked = Property(int, _get_checked, _set_checked, notify=checkedChanged)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def create_button(self, data_dict):
        button_theme = self.theme()

        btn = MenuTabBlockButton()
        if data_dict.get('image'):
            btn.image(data_dict.get('image'))
        if data_dict.get('text'):
            if data_dict.get('image') or data_dict.get('icon'):
                btn.text_beside_icon()
            else:
                btn.text_only()
        btn.theme_size = button_theme.large if button_theme else theme.Theme.Sizes.LARGE

        return btn


class MenuTabWidget(base.BaseWidget, object):
    def __init__(self, parent=None):
        super(MenuTabWidget, self).__init__(parent=parent)

    def get_main_layout(self):
        main_layout = QVBoxLayout()
        main_layout.setSpacing(0)
        main_layout.setContentsMargins(0, 0, 0, 0)

        return main_layout

    def ui(self):
        super(MenuTabWidget, self).ui()

        self.tool_btn_grp = MenuTabBlockButtonGroup()

        bar_widget = QWidget()
        bar_widget.setObjectName('bar_widget')
        self._bar_layout = QHBoxLayout()
        self._bar_layout.setContentsMargins(10, 0, 10, 0)
        bar_widget.setLayout(self._bar_layout)
        self._bar_layout.addWidget(self.tool_btn_grp)
        self._bar_layout.addStretch()

        self.main_layout.addWidget(bar_widget)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addSpacing(5)

    def append_widget(self, widget):
        """
        Adds a new widget to the end of the toolbar
        :param widget: QWidget
        """

        self._bar_layout.addWidget(widget)

    def insert_widget(self, widget):
        """
        Inserts the widget to the start of the toolbar
        :param widget: QWidget
        """

        self._bar_layout.insertWidget(0, widget)

    def add_menu(self, data_dict, index=None):
        """
        Adds a new menu
        :param data_dict: dict
        :param index: int
        """

        self.tool_btn_grp.add_button(data_dict, index)


class MenuLineButton(buttons.BaseToolButton, object):
    def __init__(self, parent=None):
        super(MenuLineButton, self).__init__(parent=parent)

        self.setCheckable(True)


class MenuLineButtonGroup(group.BaseButtonGroup, object):
    checkedChanged = Signal(int)

    def __init__(self, parent=None):
        super(MenuLineButtonGroup, self).__init__(parent=parent)

        self.set_spacing(1)
        self._button_group.setExclusive(True)
        self._button_group.buttonClicked[int].connect(self.checkedChanged)

    def create_button(self, data_dict):
        new_button = MenuLineButton(parent=self)
        if data_dict.get('image'):
            new_button.image(data_dict.get('image'))
        if data_dict.get('text'):
            if data_dict.get('image') or data_dict.get('icon'):
                new_button.text_beside_icon()
            else:
                new_button.text_only()

        return new_button

    def _get_checked(self):
        """
        Returns current checked button's ID
        :return: int
        """

        return self._button_group.checkedId()

    def _set_checked(self, value):
        """
        Sets current checked button's ID
        :param value: int
        """

        btn = self._button_group.button(value)
        btn.setChecked(True)
        self.checkedChanged.emit(value)

    theme_checked = Property(int, _get_checked, _set_checked, notify=checkedChanged)


class MenuLineTabWidget(base.BaseWidget, object):
    def __init__(self, alignment=Qt.AlignCenter, parent=None):
        self._alignment = alignment
        super(MenuLineTabWidget, self).__init__(parent=parent)

    @property
    def tool_button_group(self):
        return self._tool_button_group

    def get_main_layout(self):
        main_layout = QVBoxLayout()
        main_layout.setSpacing(0)
        main_layout.setContentsMargins(0, 0, 0, 0)

        return main_layout

    def ui(self):
        super(MenuLineTabWidget, self).ui()

        self._tool_button_group = MenuLineButtonGroup()
        self._bar_layout = QHBoxLayout()
        self._bar_layout.setContentsMargins(0, 0, 0, 0)
        if self._alignment == Qt.AlignCenter:
            self._bar_layout.addStretch()
            self._bar_layout.addWidget(self._tool_button_group)
            self._bar_layout.addStretch()
        elif self._alignment == Qt.AlignLeft:
            self._bar_layout.addWidget(self._tool_button_group)
            self._bar_layout.addStretch()
        elif self._alignment == Qt.AlignRight:
            self._bar_layout.addStretch()
            self._bar_layout.addWidget(self._tool_button_group)
        self._stack = stack.SlidingOpacityStackedWidget()
        self._tool_button_group.checkedChanged.connect(self._stack.setCurrentIndex)
        self.main_layout.addLayout(self._bar_layout)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addSpacing(5)
        self.main_layout.addWidget(self._stack)

    def add_tab(self, widget, data_dict):
        """
        Adds a new tab with given data
        :param widget: QWidget
        :param data_dict: dict
        """

        self._stack.addWidget(widget)
        self._tool_button_group.add_button(data_dict, self._stack.count() - 1)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains different types of Qt timers
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *


class ClickTimer(QObject, object):
    EXECUTE = Signal()

    def __init__(self):
        super(ClickTimer, self).__init__()
        self.timer_id = None
        self.init_data()

    def set_data(self, button, modifier, pos, selected):
        self.button = button
        self.modifier = modifier
        self.pos = pos
        self.isSelected = selected

    def init_data(self):
        self.button = None
        self.modifier = None
        self.pos = None
        self.isSelected = False

    def start(self, interval):
        self.timer_id = self.startTimer(interval)

    def remove_timer(self):
        if self.timer_id:
            self.killTimer(self.timer_id)
        self.timer_id = None
        return

    def timerEvent(self, event):
        if self.timer_id == event.timerId():
            self.EXECUTE.emit()
        self.remove_timer()


def defer(delay, fn, default_delay=1):
    """
    Append artificial delay to `func`
    This aids in keeping the GUI responsive, but complicates logic
    when producing tests. To combat this, the environment variable ensures
    that every operation is synchronous.
    :param delay: float, Delay multiplier; default 1, 0 means no delay
    :param fn: callable, Any callable
    :param default_delay: float
    """

    delay *= float(default_delay)
    if delay > 0:
        return QTimer.singleShot(delay, fn)
    else:
        return fn()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains theme implementation
"""

from __future__ import print_function, division, absolute_import

import os

from Qt.QtCore import *
from Qt.QtGui import *

from six import string_types

import tpDcc
from tpDcc.libs import qt
from tpDcc.libs.python import yamlio, color, python
from tpDcc.libs.qt.core import style, qtutils, cache, color as qt_color


class Theme(QObject, object):

    class Sizes(object):

        TINY = 18
        SMALL = 24
        MEDIUM = 32
        LARGE = 40
        HUGE = 48

    class Colors(object):

        BLUE = '#1890FF'
        PURPLE = '#722ED1'
        CYAN = '#13C2C2'
        GREEN = '#52C41A'
        MAGENTA = '#EB2F96'
        PINK = '#EF5B97'
        RED = '#F5222D'
        ORANGE = '#FA8C16'
        YELLOW = '#FADB14'
        VOLCANO = '#FA541C'
        GEEK_BLUE = '#2F54EB'
        LIME = '#A0D911'
        GOLD = '#FAAD14'

    updated = Signal()

    EXTENSION = 'yml'
    DEFAULT_ACCENT_COLOR = QColor(0, 175, 255)
    DEFAULT_SIZE = Sizes.SMALL

    def __init__(self, theme_file=None, accent_color=None, dark_mode=True):
        super(Theme, self).__init__()

        self._name = 'Default'
        self._style = 'default'
        self._file = theme_file
        self._dpi = 1
        self._background_color = None
        self._overrides = list()

        self._init_colors()
        self._init_sizes()
        self._init_font()

        accent_color = accent_color or self.Colors.BLUE
        if dark_mode:
            self.set_dark(accent_color)
        else:
            self.set_light(accent_color)

        self._init_icons()

        self.unit = 'px'
        self.default_size = self.small
        self.text_color_inverse = '#FFF'

        self._load_theme_data_from_file(theme_file)

    def __getattr__(self, item):
        options = self.options(skip_instance_attrs=True)
        if not options or item not in options:
            return super(Theme, self).__getattribute__(item)

        option_value = options[item]
        if isinstance(option_value, string_types):
            if option_value.startswith('^'):
                return qtutils.dpi_scale(int(option_value[1:]))
            if color.string_is_hex(option_value):
                return color.hex_to_rgb(option_value)
        else:
            return option_value

    def _load_theme_data_from_file(self, theme_file):
        """
        Internal function that laods file data from given file
        :param theme_file: str
        :return: dict
        """

        if not theme_file or not os.path.isfile(theme_file):
            return

        try:
            theme_data = yamlio.read_file(theme_file)
        except Exception:
            qt.logger.warning('Impossible to load theme data from file: "{}"!'.format(theme_file))
            return None

        theme_name = theme_data.get('name', None)
        if not theme_name:
            qt.logger.warning('Impossible to retrieve them name from theme file: "{}"!'.format(theme_file))
            return None
        accent_color = theme_data.get('accent_color', None)
        if not accent_color:
            qt.logger.warning('No theme color definitions found in theme file: "{}"'.format(theme_file))
            return None
        self._style = theme_data.get('style', 'default.css')
        self._overrides = theme_data.get('overrides', list())

        self.set_name(theme_name)
        self.set_accent_color(accent_color)

    def name(self):
        """
        Returns the name for this theme
        :return: str
        """

        return self._name

    def set_name(self, name):
        """
        Sets the name for this theme
        :param name: str
        """

        self._name = name

    def dpi(self):
        """
        Returns zoom amount for this theme
        :return: float
        """

        return self._dpi

    def set_dpi(self, dpi):
        """
        Sets the zoom amount for this theme
        :param dpi: float
        """

        self._dpi = dpi

    def set_accent_color(self, accent_color):
        """
        Sets the main/accent color of the theme
        :param accent_color:
        """

        self._update_accent_color(accent_color)
        self.updated.emit()

    def is_dark(self):
        """
        Returns whether the current theme is dark or not
        :return: bool
        """

        bg_color = qt_color.Color(self.background_color)

        red = bg_color.redF() * 0.299
        green = bg_color.greenF() * 0.587
        blue = bg_color.blueF() * 0.114

        darkness = red + green + blue
        if darkness < 0.6:
            return True

        return False

    def set_dark(self, accent_color):
        """
        Sets the current theme to the default dark color
        """

        self.background_color = '#323232'
        self.background_selected_color = '#292929'
        self.background_in_color = '#3A3A3A'
        self.background_out_color = '#494949'
        self.mask_color = self._fade_color(self.background_color, '90%')
        self.toast_color = '#555555'
        self.title_color = "#FFFFFF"
        self.primary_text_color = "#D9D9D9"
        self.secondary_text_color = "#A6A6A6"
        self.disable_color = "#737373"
        self.border_color = "#1E1E1E"
        self.divider_color = "#262626"
        self.header_color = "#0A0A0A"
        self.icon_color = "#A6A6A6"
        self.window_dragger_color = "#232323"
        self.window_dragger_label_color = "#D9D9D9"

        self.set_accent_color(accent_color)

    def set_light(self, accent_color):
        """
        Sets the current theme to the default light color
        """

        self.background_color = '#F8F8F9'
        self.background_selected_color = '#BFBFBF'
        self.background_in_color = '#FFFFFF'
        self.background_out_color = '#EEEEEE'
        self.mask_color = self._fade_color(self.background_color, '90%')
        self.toast_color = '#333333'
        self.title_color = "#262626"
        self.primary_text_color = "#595959"
        self.secondary_text_color = "#8C8C8C"
        self.disable_color = "#E5E5E5"
        self.border_color = "#D9D9D9"
        self.divider_color = "#E8E8E8"
        self.header_color = "#FAFAFA"
        self.icon_color = "#8C8C8C"
        self.window_dragger_color = "#f2f2fd"
        self.window_dragger_label_color = "#262626"

        self.set_accent_color(accent_color)

    def _init_colors(self):
        """
        Internal function that initializes all theme colors
        """

        self.info_color = self.Colors.BLUE
        self.success_color = self.Colors.GREEN
        self.processing_color = self.Colors.BLUE
        self.error_color = self.Colors.RED
        self.warning_color = self.Colors.GOLD

        self.info_1 = self._fade_color(self.info_color, '15%')
        self.info_2 = qt_color.generate_color(self.info_color, 2)
        self.info_3 = self._fade_color(self.info_color, '35%')
        self.info_4 = qt_color.generate_color(self.info_color, 4)
        self.info_5 = qt_color.generate_color(self.info_color, 5)
        self.info_6 = qt_color.generate_color(self.info_color, 6)
        self.info_7 = qt_color.generate_color(self.info_color, 7)
        self.info_8 = qt_color.generate_color(self.info_color, 8)
        self.info_9 = qt_color.generate_color(self.info_color, 9)
        self.info_10 = qt_color.generate_color(self.info_color, 10)

        self.success_1 = self._fade_color(self.success_color, '15%')
        self.success_2 = qt_color.generate_color(self.success_color, 2)
        self.success_3 = self._fade_color(self.success_color, '35%')
        self.success_4 = qt_color.generate_color(self.success_color, 4)
        self.success_5 = qt_color.generate_color(self.success_color, 5)
        self.success_6 = qt_color.generate_color(self.success_color, 6)
        self.success_7 = qt_color.generate_color(self.success_color, 7)
        self.success_8 = qt_color.generate_color(self.success_color, 8)
        self.success_9 = qt_color.generate_color(self.success_color, 9)
        self.success_10 = qt_color.generate_color(self.success_color, 10)

        self.warning_1 = self._fade_color(self.warning_color, '15%')
        self.warning_2 = qt_color.generate_color(self.warning_color, 2)
        self.warning_3 = self._fade_color(self.warning_color, '35%')
        self.warning_4 = qt_color.generate_color(self.warning_color, 4)
        self.warning_5 = qt_color.generate_color(self.warning_color, 5)
        self.warning_6 = qt_color.generate_color(self.warning_color, 6)
        self.warning_7 = qt_color.generate_color(self.warning_color, 7)
        self.warning_8 = qt_color.generate_color(self.warning_color, 8)
        self.warning_9 = qt_color.generate_color(self.warning_color, 9)
        self.warning_10 = qt_color.generate_color(self.warning_color, 10)

        self.error_1 = self._fade_color(self.error_color, '15%')
        self.error_2 = qt_color.generate_color(self.error_color, 2)
        self.error_3 = self._fade_color(self.error_color, '35%')
        self.error_4 = qt_color.generate_color(self.error_color, 4)
        self.error_5 = qt_color.generate_color(self.error_color, 5)
        self.error_6 = qt_color.generate_color(self.error_color, 6)
        self.error_7 = qt_color.generate_color(self.error_color, 7)
        self.error_8 = qt_color.generate_color(self.error_color, 8)
        self.error_9 = qt_color.generate_color(self.error_color, 9)
        self.error_10 = qt_color.generate_color(self.error_color, 10)

    def _init_sizes(self):
        """
        Internal function that initializes all themes sizes
        """

        self.border_radius_large = 8
        self.border_radius_base = 4
        self.border_radius_small = 2
        self.tiny = self.Sizes.TINY
        self.small = self.Sizes.SMALL
        self.medium = self.Sizes.MEDIUM
        self.large = self.Sizes.LARGE
        self.huge = self.Sizes.HUGE
        self.tiny_icon = self.tiny - 8
        self.small_icon = self.small - 10
        self.medium_icon = self.medium - 12
        self.large_icon = self.large - 16
        self.huge_icon = self.huge - 20
        self.window_dragger_rounded_corners = 5
        self.window_dragger_font_size = 12
        self.window_rounded_corners = 5
        self.button_padding = 4

    def _init_font(self):
        """
        Internal function that initializes all theme fonts
        """

        self.font_family = 'BlinkMacSystemFont,"Segoe UI","PingFang SC","Hiragino Sans GB","Microsoft YaHei",' \
                           '"Helvetica Neue",Helvetica,Arial,sans-serif'
        self.font_size_base = 14
        self.font_size_large = self.font_size_base + 2
        self.font_size_small = self.font_size_base - 2
        self.h1_size = int(self.font_size_base * 2.71)
        self.h2_size = int(self.font_size_base * 2.12)
        self.h3_size = int(self.font_size_base * 1.71)
        self.h4_size = int(self.font_size_base * 1.41)

    def _init_icons(self):
        """
        Internal function that initializes all theme icons
        """

        self.radio_checked_icon = 'radio_button_checked.png'
        self.radio_unchecked_icon = 'radio_button_unchecked.png'
        self.up_icon = 'collapse.png'
        self.down_icon = 'expand.png'
        self.up_arrow_icon = 'up_arrow.png'
        self.down_arrow_icon = 'down_arrow.png'
        self.left_icon = 'back.png'
        self.right_icon = 'next.png'
        self.calendar_icon = 'calendar.png'
        self.check_icon = 'check.png'
        self.uncheck_icon = 'uncheck.png'
        self.circle_icon = 'circle.png'
        self.splitter_icon = 'splitter.png'

    def _update_accent_color(self, accent_color):
        accent_color = qt_color.convert_2_hex(accent_color)
        self.accent_color = accent_color
        self.accent_color_1 = qt_color.generate_color(accent_color, 1)
        self.accent_color_2 = qt_color.generate_color(accent_color, 2)
        self.accent_color_3 = qt_color.generate_color(accent_color, 3)
        self.accent_color_4 = qt_color.generate_color(accent_color, 4)
        self.accent_color_5 = qt_color.generate_color(accent_color, 5)
        self.accent_color_6 = qt_color.generate_color(accent_color, 6)
        self.accent_color_7 = qt_color.generate_color(accent_color, 7)
        self.accent_color_8 = qt_color.generate_color(accent_color, 8)
        self.accent_color_9 = qt_color.generate_color(accent_color, 9)
        self.accent_color_10 = qt_color.generate_color(accent_color, 10)
        self.item_hover_background_color = self.accent_color_1

    def _get_color(self, color_value, alpha=None):
        """
        Internal function that returns a color value in proper format to be handled by theme
        :param color_value: variant, str or QColor or color.Color
        """

        if isinstance(color_value, (str, unicode)):
            color_value = qt_color.Color.from_string(color_value)
        elif isinstance(color_value, QColor):
            color_value = qt_color.Color.from_color(color_value)
        elif isinstance(color_value, (list, tuple)):
            color_value = qt_color.Color(*color_value)

        return color_value

    def _fade_color(self, color, alpha):
        """
        Internal function that fades given color
        :param color: QColor
        :param alpha: float
        :return:
        """

        qt_color = QColor(color)
        return 'rgba({}, {}, {}, {})'.format(qt_color.red(), qt_color.green(), qt_color.blue(), alpha)

    def foreground_color(self):
        """
        Returns the foreground color for this theme
        :return: color.Color
        """

        if self.is_dark():
            return qt_color.Color(250, 250, 250, 255)
        else:
            return qt_color.Color(0, 40, 80, 180)

    # def icon_color(self):
    #     """
    #     Returns the icon color for this theme
    #     :return: color.Color
    #     """
    #
    #     return self.foreground_color()
    #
    # def accent_foreground_color(self):
    #     """
    #     Returns the foregound color for the accent color
    #     """
    #
    #     return qt_color.Color(255, 255, 255, 255)
    #
    # def item_background_color(self):
    #     """
    #     Returns the item background color
    #     :return: color.Color
    #     """
    #
    #     if self.is_dark():
    #         return qt_color.Color(255, 255, 255, 20)
    #     else:
    #         return qt_color.Color(255, 255, 255, 120)
    #
    # def item_background_hover_color(self):
    #     """
    #     Returns the item background color when the mouse hovers over the item
    #     :return: color.Color
    #     """
    #
    #     return qt_color.Color(255, 255, 255, 60)
    #
    # def settings(self):
    #     """
    #     Returns a dictionary of settings for the current theme
    #     :return: dict
    #     """
    #
    #     return {
    #         'name': self.name(),
    #         'accentColor': self.accent_color().to_string(),
    #         'backgroundColor': self.background_color().to_string()
    #     }

    def set_settings(self, settings):
        """
        Sets a dictionary of settings for the current theme
        :param settings: dict
        """

        for theme_sett_name, theme_sett_value in settings.items():
            if hasattr(self, theme_sett_name):
                setattr(self, theme_sett_name, theme_sett_value)

    def get_theme_option(self, option_name, default_value=None):
        """
        Returns option of the style
        :return: object
        """

        theme_options = self.options()
        if not theme_options:
            return default_value

        return theme_options[option_name] if option_name in theme_options else default_value

    def options(self, skip_instance_attrs=False):
        """
        Returns the variables used to customize the style sheet
        :return: dict
        """
        if self.is_dark():
            darkness = 'white'
        else:
            darkness = 'black'

        theme_resources_dir = ''
        if self._file and os.path.isfile(self._file):
            theme_dir = os.path.dirname(self._file)
            theme_name = os.path.splitext(os.path.basename(self._file))[0]
            theme_resources_dir = os.path.join(theme_dir, 'resources', theme_name)

        style_resources_dir = ''
        style_path = self.stylesheet_file()
        if style_path and os.path.isfile(style_path):
            style_dir = os.path.dirname(style_path)
            style_name = os.path.splitext(os.path.basename(style_path))[0]
            style_resources_dir = os.path.join(style_dir, 'resources', style_name)

        options = {
            'darkness': darkness,
            'theme_resources': theme_resources_dir,
            'style_resources': style_resources_dir
        }

        if not skip_instance_attrs:
            inst_attrs = python.get_instance_user_attributes(self)
            for attr in inst_attrs:
                if isinstance(attr[1], QColor):
                    options[attr[0]] = qt_color.Color(attr[1]).to_string()
                else:
                    options[attr[0]] = str(attr[1])

    #     options = {
    #         "ACCENT_COLOR": accent_color.to_string(),
    #         "ACCENT_COLOR_DARKER": qt_color.Color(accent_color.darker(150)).to_string(),
    #         "ACCENT_COLOR_LIGHTER": qt_color.Color(accent_color.lighter(150)).to_string(),
    #         "ACCENT_COLOR_R": str(accent_color.red()),
    #         "ACCENT_COLOR_G": str(accent_color.green()),
    #         "ACCENT_COLOR_B": str(accent_color.blue()),
    #
    #         "ACCENT_FOREGROUND_COLOR": accent_foreground_color.to_string(),
    #         "ACCENT_FOREGROUND_COLOR_DARKER": qt_color.Color(accent_foreground_color.darker(150)).to_string(),
    #
    #         "FOREGROUND_COLOR": foreground_color.to_string(),
    #         "FOREGROUND_COLOR_R": str(foreground_color.red()),
    #         "FOREGROUND_COLOR_G": str(foreground_color.green()),
    #         "FOREGROUND_COLOR_B": str(foreground_color.blue()),
    #
    #         "BACKGROUND_COLOR": background_color.to_string(),
    #         "BACKGROUND_COLOR_LIGHTER": qt_color.Color(background_color.lighter(150)).to_string(),
    #         "BACKGROUND_COLOR_DARKER": qt_color.Color(background_color.darker(150)).to_string(),
    #         "BACKGROUND_COLOR_R": str(background_color.red()),
    #         "BACKGROUND_COLOR_G": str(background_color.green()),
    #         "BACKGROUND_COLOR_B": str(background_color.blue()),
    #
    #         "ITEM_TEXT_COLOR": foreground_color.to_string(),
    #         "ITEM_TEXT_SELECTED_COLOR": accent_foreground_color.to_string(),
    #
    #         "ITEM_BACKGROUND_COLOR": item_background_color.to_string(),
    #         "ITEM_BACKGROUND_HOVER_COLOR": item_background_hover_color.to_string(),
    #         "ITEM_BACKGROUND_SELECTED_COLOR": accent_color.to_string(),
    #     }
    #
        overrides = self._overrides or dict()
        options.update(overrides)
    #
        return options

    def stylesheet_file(self):
        """
        Returns path where theme stylesheet is located
        :return: str
        """

        style_name = self._style or 'default'
        style_extension = style.StyleSheet.EXTENSION
        if not style_extension.startswith('.'):
            style_extension = '.{}'.format(style_extension)
        style_file_name = '{}{}'.format(style_name, style_extension)
        style_path = tpDcc.ResourcesMgr().get('styles', style_file_name)

        return style_path

    def stylesheet(self):
        """
        Returns the style sheet for this theme
        :return: str
        """

        style_path = self.stylesheet_file()
        options = self.options()

        stylesheet = style.StyleSheet.from_path(style_path, options=options, theme_name=self._name, dpi=self.dpi())

        return stylesheet.data()
    #
    # def create_color_dialog(self, parent, standard_colors=None, current_color=None):
    #     """
    #     Creates a new instance of color dialog
    #     :param parent: QWidget
    #     :param standard_colors: list(int)
    #     :param current_color: QColor
    #     :return: QColorDialog
    #     """
    #
    #     dlg = QColorDialog(parent)
    #     if standard_colors:
    #         index = -1
    #         for r, g, b in standard_colors:
    #             index += 1
    #             clr = QColor(r, g, b).rgba()
    #             try:
    #                 clr = QColor(clr)
    #                 dlg.setStandardColor(index, clr)
    #             except Exception:
    #                 clr = QColor(clr).rgba()
    #                 dlg.setStandardColor(index, clr)
    #
    #     # PySide2 does not supports d.open(), we pass a blank slot
    #     dlg.open(self, Slot('blankSlot()'))
    #
    #     if current_color:
    #         dlg.setCurrentColor(current_color)
    #
    #     return dlg
    #
    # def browse_accent_color(self, parent=None):
    #     """
    #     Shows the color dialog for changing the accent color
    #     :param parent: QWidget
    #     """
    #
    #     standard_colors = [
    #         (230, 60, 60), (210, 40, 40), (190, 20, 20), (250, 80, 130),
    #         (230, 60, 110), (210, 40, 90), (255, 90, 40), (235, 70, 20),
    #         (215, 50, 0), (240, 100, 170), (220, 80, 150), (200, 60, 130),
    #         (255, 125, 100), (235, 105, 80), (215, 85, 60), (240, 200, 150),
    #         (220, 180, 130), (200, 160, 110), (250, 200, 0), (230, 180, 0),
    #         (210, 160, 0), (225, 200, 40), (205, 180, 20), (185, 160, 0),
    #         (80, 200, 140), (60, 180, 120), (40, 160, 100), (80, 225, 120),
    #         (60, 205, 100), (40, 185, 80), (50, 180, 240), (30, 160, 220),
    #         (10, 140, 200), (100, 200, 245), (80, 180, 225), (60, 160, 205),
    #         (130, 110, 240), (110, 90, 220), (90, 70, 200), (180, 160, 255),
    #         (160, 140, 235), (140, 120, 215), (180, 110, 240), (160, 90, 220),
    #         (140, 70, 200), (210, 110, 255), (190, 90, 235), (170, 70, 215)
    #     ]
    #
    #     current_color = self.accent_color()
    #
    #     dialog = self.create_color_dialog(parent, standard_colors, current_color)
    #     dialog.currentColorChanged.connect(self.set_accent_color)
    #
    #     if dialog.exec_():
    #         self.set_accent_color(dialog.selectedColor())
    #     else:
    #         self.set_accent_color(current_color)
    #
    # def browse_background_color(self, parent=None):
    #     """
    #     Shows the color dialog for changing the background color
    #     :param parent: QWidget
    #     """
    #
    #     standard_colors = [
    #         (0, 0, 0), (20, 20, 20), (40, 40, 40), (60, 60, 60),
    #         (80, 80, 80), (100, 100, 100), (20, 20, 30), (40, 40, 50),
    #         (60, 60, 70), (80, 80, 90), (100, 100, 110), (120, 120, 130),
    #         (0, 30, 60), (20, 50, 80), (40, 70, 100), (60, 90, 120),
    #         (80, 110, 140), (100, 130, 160), (0, 60, 60), (20, 80, 80),
    #         (40, 100, 100), (60, 120, 120), (80, 140, 140), (100, 160, 160),
    #         (0, 60, 30), (20, 80, 50), (40, 100, 70), (60, 120, 90),
    #         (80, 140, 110), (100, 160, 130), (60, 0, 10), (80, 20, 30),
    #         (100, 40, 50), (120, 60, 70), (140, 80, 90), (160, 100, 110),
    #         (60, 0, 40), (80, 20, 60), (100, 40, 80), (120, 60, 100),
    #         (140, 80, 120), (160, 100, 140), (40, 15, 5), (60, 35, 25),
    #         (80, 55, 45), (100, 75, 65), (120, 95, 85), (140, 115, 105)
    #     ]
    #
    #     current_color = self.background_color()
    #
    #     dialog = self.create_color_dialog(parent, standard_colors, current_color)
    #     dialog.currentColorChanged.connect(self.set_background_color)
    #
    #     if dialog.exec_():
    #         self.set_background_color(dialog.selectedColor())
    #     else:
    #         self.set_background_color(current_color)


ThemeCache = cache.CacheResource(Theme)



# This file is part of the PySide project.
#
# Copyright (C) 2009-2011 Nokia Corporation and/or its subsidiary(-ies).
# Copyright (C) 2010 Riverbank Computing Limited.
# Copyright (C) 2009 Torsten Marek
#
# Contact: PySide team <pyside@openbossa.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301 USA


import sys
import logging

from pysideuic import compileUi


class Driver(object):
    """ This encapsulates access to the pyuic functionality so that it can be
    called by code that is Python v2/v3 specific.
    """

    LOGGER_NAME = 'PySide.uic'

    def __init__(self, opts, ui_file):
        """ Initialise the object.  opts is the parsed options.  ui_file is the
        name of the .ui file.
        """

        if opts.debug:
            logger = logging.getLogger(self.LOGGER_NAME)
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter("%(name)s: %(message)s"))
            logger.addHandler(handler)
            logger.setLevel(logging.DEBUG)

        self._opts = opts
        self._ui_file = ui_file

    def invoke(self):
        """ Invoke the action as specified by the parsed options.  Returns 0 if
        there was no error.
        """

        if self._opts.preview:
            return self._preview()

        self._generate()

        return 0

    def _preview(self):
        """ Preview the .ui file.  Return the exit status to be passed back to
        the parent process.
        """

        from PySide import QtUiTools
        from PySide import QtGui

        app = QtGui.QApplication([self._ui_file])
        widget = QtUiTools.QUiLoader().load(self._ui_file)
        widget.show()

        return app.exec_()

    def _generate(self):
        """ Generate the Python code. """

        if sys.hexversion >= 0x03000000:
            if self._opts.output == '-':
                from io import TextIOWrapper

                pyfile = TextIOWrapper(sys.stdout.buffer, encoding='utf8')
            else:
                pyfile = open(self._opts.output, 'wt', encoding='utf8')
        else:
            if self._opts.output == '-':
                pyfile = sys.stdout
            else:
                pyfile = open(self._opts.output, 'wt')

        compileUi(self._ui_file, pyfile, self._opts.execute, self._opts.indent, self._opts.from_imports)

    def on_IOError(self, e):
        """ Handle an IOError exception. """

        sys.stderr.write("Error: %s: \"%s\"\n" % (e.strerror, e.filename))

    def on_SyntaxError(self, e):
        """ Handle a SyntaxError exception. """

        sys.stderr.write("Error in input file: %s\n" % e)

    def on_NoSuchWidgetError(self, e):
        """ Handle a NoSuchWidgetError exception. """

        if e.args[0].startswith("Q3"):
            sys.stderr.write("Error: Q3Support widgets are not supported by PySide.\n")
        else:
            sys.stderr.write(str(e) + "\n")

    def on_Exception(self, e):
        """ Handle a generic exception. """

        if logging.getLogger(self.LOGGER_NAME).level == logging.DEBUG:
            import traceback

            traceback.print_exception(*sys.exc_info())
        else:
            from PySide import QtCore

            sys.stderr.write("""An unexpected error occurred.
Check that you are using the latest version of PySide and report the error to
http://bugs.openbossa.org, including the ui file used to trigger the error.
""")



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Collapsible accordion widget similar to Maya Attribute Editor
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *


class BalloonDialog(QDialog, object):

    FIXED_HEIGHT = 15
    FIXED_WIDTH = 35

    class BallonDialogFocuser(QWidget, object):
        def __init__(self, w):
            super(BalloonDialog.BallonDialogFocuser, self).__init__()

            self._widget = w

        def show(self):
            """
            Overrides base QWidget show function
            """

            self._widget.show()
            self.focus()

        def focus(self):
            """
            Function that focus wrapped widget
            """

            self._widget.activateWindow()
            self._widget.raise_()

    def __init__(self, modal=False, parent=None):
        super(BalloonDialog, self).__init__(parent)

        self._lazy_show_window = QTimer(self)

        if modal:
            self.setModal(True)

        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)

        self._lazy_show_window.timeout.connect(self._on_lazy_show_window)

    def resizeEvent(self, event):
        """
        Overrides base QDialog resizeEvent function
        :param event: QResizeEvent
        """

        r = self.rect()
        ss = self.styleSheet()
        # ss.replace(str(QRegExp("\\s*margin-top\\s*:\\s*.+?;")), "")
        self.setStyleSheet('{} margin-top: {}px;'.format(ss, self.FIXED_HEIGHT))

        poly = QPolygon()
        poly.append(QPoint(r.x(), r.y() + self.FIXED_HEIGHT))
        poly.append(QPoint(r.x() + r.width() / 2 - self.FIXED_WIDTH / 2, r.y() + self.FIXED_HEIGHT))
        poly.append(QPoint(r.x() + r.width() / 2, r.y()))
        poly.append(QPoint(r.x() + r.width() / 2 + self.FIXED_WIDTH / 2, r.y() + self.FIXED_HEIGHT))
        poly.append(QPoint(r.x() + r.width(), r.y() + self.FIXED_HEIGHT))
        poly.append(QPoint(r.x() + r.width(), r.y() + r.height()))
        poly.append(QPoint(r.x(), r.y() + r.height()))

        new_mask = QRegion(poly)
        self.setMask(new_mask)

    def event(self, e):
        """
        Overrides base QDialog event function
        :param e: QEvent
        :return: bool
        """

        if QEvent.WindowDeactivate == e.type():
            self.done(False)
            e.ignore()
            return True

        return super(BalloonDialog, self).event(e)

    def focusOutEvent(self, event):
        """
        Overrides base QDialog focusOutEvent function
        :param event: QFocusEven
        """

        pass

    def showEvent(self, event):
        """
        Overrides base QDialog showEvent function
        :param event: QShowEvent
        """

        focuser = BalloonDialog.BallonDialogFocuser(self)
        focuser.focus()

    def _on_lazy_show_window(self):
        """
        Internal callback function that is called when lazy show window timer finishes
        """

        self.activateWindow()
        self.setFocus()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains classes to create different kind of joystick widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *
from Qt.QtGui import *


class JoyPad(QWidget, object):

    xChanged = Signal(float)
    yChanged = Signal(float)

    def __init__(self, parent=None):
        super(JoyPad, self).__init__(parent=parent)

        self._x = 0
        self._y = 0
        self._bounds = QRectF()
        self._knop_bounds = QRectF()
        self._last_pos = QPoint()
        self._knop_pressed = False

        self._return_animation = QParallelAnimationGroup(self)
        self._x_anim = QPropertyAnimation(self, 'x')
        self._y_anim = QPropertyAnimation(self, 'y')
        self._alignment = Qt.AlignTop | Qt.AlignLeft

        self._x_anim.setEndValue(0.0)
        self._x_anim.setDuration(400)
        self._x_anim.setEasingCurve(QEasingCurve.OutSine)

        self._y_anim.setEndValue(0.0)
        self._y_anim.setDuration(400)
        self._y_anim.setEasingCurve(QEasingCurve.OutSine)

        self._return_animation.addAnimation(self._x_anim)
        self._return_animation.addAnimation(self._y_anim)

    # region Static Functions
    @staticmethod
    def constraint(value, min_value, max_value):
        return min_value if value < min_value else max_value if value > max_value else value
    # endregion

    # region Properties
    def get_x(self):
        return self._x

    def set_x(self, x):
        self._x = self.constraint(x, -1.0, 1.0)
        radius = (self._bounds.width() - self._knop_bounds.width()) * 0.5
        self._knop_bounds.moveCenter(
            QPointF(self._bounds.center().x() + self._x * radius, self._knop_bounds.center().y()))
        self.update()
        self.xChanged.emit(self._x)

    def get_y(self):
        return self._y

    def set_y(self, y):
        self._y = self.constraint(y, -1.0, 1.0)
        radius = (self._bounds.width() - self._knop_bounds.width()) * 0.5
        self._knop_bounds.moveCenter(
            QPointF(self._knop_bounds.center().x(), self._bounds.center().y() - self._y * radius))
        self.update()
        self.yChanged.emit(self._x)

    def get_alignment(self):
        return self._alignment

    def set_alignment(self, alignment):
        self._alignment = alignment

    x = property(get_x, set_x)
    y = property(get_y, set_y)
    alignment = property(get_alignment, set_alignment)
    # endregion

    # region Override Functions
    def resizeEvent(self, event):
        a = min(self.width(), self.height())
        top_left = QPointF()
        if self._alignment & Qt.AlignTop:
            top_left.setY(0)
        elif self._alignment & Qt.AlignVCenter:
            top_left.setY((self.height() - a) * 0.5)
        elif self._alignment & Qt.AlignBottom:
            top_left.setY(self.height() - a)

        if self._alignment & Qt.AlignLeft:
            top_left.setX(0)
        elif self._alignment & Qt.AlignHCenter:
            top_left.setX((self.width() - a) * 0.5)
        elif self._alignment & Qt.AlignRight:
            top_left.setX(self.width() - a)

        self._bounds = QRectF(top_left, QSize(a, a))
        self._knop_bounds.setWidth(a * 0.3)
        self._knop_bounds.setHeight(a * 0.3)

        radius = (self._bounds.width() - self._knop_bounds.height()) * 0.5
        self._knop_bounds.moveCenter(
            QPointF(self._bounds.center().x() + self._x * radius, self._bounds.center().y() - self._y * radius))

    def mousePressEvent(self, event):
        if self._knop_bounds.contains(event.pos()):
            # self._return_animation.stop()
            self._last_pos = event.pos()
            self._knop_pressed = True

    def mouseReleaseEvent(self, event):
        self._knop_pressed = False
        self.x = 0.0
        self.y = 0.0
        # self._return_animation.start()

    def mouseMoveEvent(self, event):
        if not self._knop_pressed:
            return

        delta_pos = QPointF(event.pos() - self._last_pos)
        delta_pos += 0.5 * (QPointF(event.pos()) - self._knop_bounds.center())

        from_center_to_knop = self._knop_bounds.center() + delta_pos - self._bounds.center()
        radius = (self._bounds.width() - self._knop_bounds.width()) * 0.5
        from_center_to_knop.setX(self.constraint(from_center_to_knop.x(), -radius, radius))
        from_center_to_knop.setY(self.constraint(from_center_to_knop.y(), -radius, radius))
        self._knop_bounds.moveCenter(from_center_to_knop + self._bounds.center())
        self._last_pos = event.pos()

        self.update()

        if radius == 0:
            return
        x = (self._knop_bounds.center().x() - self._bounds.center().x()) / radius
        y = (-self._knop_bounds.center().y() + self._bounds.center().y()) / radius

        if self._x != x:
            self._x = x
            self.xChanged.emit(self._x)

        if self._y != y:
            self._y = y
            self.yChanged.emit(self._y)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setRenderHint(QPainter.HighQualityAntialiasing)

        gradient = QRadialGradient(self._bounds.center(), self._bounds.width() * 0.5, self._bounds.center())
        gradient.setFocalRadius(self._bounds.width() * 0.3)
        gradient.setCenterRadius(self._bounds.width() * 0.7)
        gradient.setColorAt(0, Qt.white)
        gradient.setColorAt(1, Qt.lightGray)

        painter.setPen(QPen(QBrush(Qt.gray), self._bounds.width() * 0.005))
        painter.setBrush(QBrush(gradient))
        painter.drawEllipse(self._bounds)

        painter.setPen(QPen(QBrush(Qt.gray), self._bounds.width() * 0.005))
        painter.drawLine(
            QPointF(self._bounds.left(), self._bounds.center().y()),
            QPointF(self._bounds.center().x() - self._bounds.width() * 0.35, self._bounds.center().y())
        )
        painter.drawLine(
            QPointF(self._bounds.center().x() + self._bounds.width() * 0.35, self._bounds.center().y()),
            QPointF(self._bounds.right(), self._bounds.center().y())
        )
        painter.drawLine(
            QPointF(self._bounds.center().x(), self._bounds.top()),
            QPointF(self._bounds.center().x(), self._bounds.center().y() - self._bounds.width() * 0.35)
        )
        painter.drawLine(
            QPointF(self._bounds.center().x(), self._bounds.center().y() + self._bounds.width() * 0.35),
            QPointF(self._bounds.center().x(), self._bounds.bottom())
        )

        if not self.isEnabled():
            return

        gradient = QRadialGradient(
            self._knop_bounds.center(), self._knop_bounds.width() * 0.5, self._knop_bounds.center())
        gradient.setFocalRadius(self._knop_bounds.width() * 0.2)
        gradient.setCenterRadius(self._knop_bounds.width() * 0.5)
        gradient.setColorAt(0, Qt.gray)
        gradient.setColorAt(1, Qt.darkGray)

        painter.setPen(QPen(QBrush(Qt.darkGray), self._bounds.width() * 0.005))
        painter.setBrush(QBrush(gradient))
        painter.drawEllipse(self._knop_bounds)

        # endregion

    # region Public Functions
    def add_x_animation(self):
        # Abort if the animation is already added
        if self._x_anim.parent() == self._return_animation:
            return

        self._return_animation.addAnimation(self._x_anim)

    def remove_x_animation(self):
        # Abort if the animation is already removed
        if self._x_anim.parent() != self._return_animation:
            return

        self._return_animation.removeAnimation(self._x_anim)

        # Take ownership of the animation (parent is 0 after removeAnimation())
        self._x_anim.setParent(self)

    def add_y_animation(self):
        # Abort if the animation is already added
        if self._y_anim.parent() == self._return_animation:
            return

        self._return_animation.addAnimation(self._y_anim)

    def remove_y_animation(self):
        # Abort if the animation is already removed
        if self._y_anim.parent() != self._return_animation:
            return

        self._return_animation.removeAnimation(self._y_anim)

        # Take ownership of the animation (parent is 0 after removeAnimation())
        self._y_anim.setParent(self)
    # endregion



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for panel widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

import tpDcc as tp
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import label, buttons, dividers


class SliderPanelPositions(object):

    LEFT = 'left'
    RIGHT = 'right'
    TOP = 'top'
    BOTTOM = 'bottom'


class SliderPanel(base.BaseWidget, object):
    """
    Panel that slides in from the edge of the window
    """

    closed = Signal()
    closeButtonClicked = Signal()

    def __init__(self, title, position=SliderPanelPositions.RIGHT, closable=True, parent=None):

        self._title = title
        self._position = position
        self._closable = closable
        self._is_first_close = True

        super(SliderPanel, self).__init__(parent)

        self.setObjectName('sliderPanel')
        self.setWindowFlags(Qt.Popup)
        self.setAttribute(Qt.WA_StyledBackground)

        self._close_timer = QTimer(self)
        self._close_timer.setInterval(300)
        self._close_timer.setSingleShot(True)
        self._close_timer.timeout.connect(self.close)
        self._close_timer.timeout.connect(self.closed.emit)

        self._pos_anim = QPropertyAnimation(self)
        self._pos_anim.setTargetObject(self)
        self._pos_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._pos_anim.setDuration(300)
        self._pos_anim.setPropertyName('pos')

        self._opacity_anim = QPropertyAnimation()
        self._opacity_anim.setTargetObject(self)
        self._opacity_anim.setDuration(300)
        self._opacity_anim.setEasingCurve(QEasingCurve.OutCubic)
        self._opacity_anim.setPropertyName('windowOpacity')
        self._opacity_anim.setStartValue(0.0)
        self._opacity_anim.setEndValue(1.0)

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    @property
    def position(self):
        """
        Returns the placement of the panel in parent window
        :return: str
        """

        return self._position

    @position.setter
    def position(self, value):
        """
        Sets the position of the panel in parent window ('top', 'right', 'bottom' or 'left').
        :param value: str
        """

        self._position = value
        if value in [SliderPanelPositions.BOTTOM, SliderPanelPositions.TOP]:
            self.setFixedHeight(200)
        else:
            self.setFixedWidth(200)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def ui(self):
        super(SliderPanel, self).ui()

        self._title_label = label.BaseLabel(parent=self).h4()
        self._title_label.setText(self._title)

        self._close_btn = buttons.BaseToolButton(parent=self).icon_only().image('close', theme='window').small()
        self._close_btn.setVisible(self._closable or False)

        title_layout = QHBoxLayout()
        title_layout.addWidget(self._title_label)
        title_layout.addStretch()
        title_layout.addWidget(self._close_btn)

        self._button_layout = QHBoxLayout()
        self._button_layout.addStretch()

        self._scroll_area = QScrollArea()
        self.main_layout.addLayout(title_layout)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addWidget(self._scroll_area)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addLayout(self._button_layout)

    def setup_signals(self):
        self._close_btn.clicked.connect(self.close)
        self._close_btn.clicked.connect(self.closeButtonClicked.emit)

    def show(self):
        self._update_position()
        self._fade_in()

        return super(SliderPanel, self).show()

    def closeEvent(self, event):
        if self._is_first_close:
            self._is_first_close = False
            self._close_timer.stop()
            self._fade_out()
            event.ignore()
        else:
            event.accept()

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def set_widget(self, widget):
        """
        Sets the widget that will be contained inside the panel
        :param widget: QWidget
        """

        self._scroll_area.setWidget(widget)

    def add_button(self, button):
        """
        Adds a new button to the bottom part of the panel
        :param button: QPushButton
        """

        self._button_layout.addWidget(button)

    def left(self):
        """
        Sets the panel's placement to left
        :return: SliderPanel
        """

        self.position = SliderPanelPositions.LEFT

        return self

    def right(self):
        """
        Sets the panel's placement to right
        :return: SliderPanel
        """

        self.position = SliderPanelPositions.RIGHT

        return self

    def top(self):
        """
        Sets the panel's placement to top
        :return: SliderPanel
        """

        self.position = SliderPanelPositions.TOP

        return self

    def bottom(self):
        """
        Sets the panel's placement to bottom
        :return: SliderPanel
        """

        self.position = SliderPanelPositions.BOTTOM

        return self

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _fade_in(self):
        """
        Internal function that fades in the panel
        """

        self._pos_anim.start()
        self._opacity_anim.start()

    def _fade_out(self):
        """
        Internal function that fades out the panel
        """

        self._pos_anim.setDirection(QAbstractAnimation.Backward)
        self._pos_anim.start()
        self._opacity_anim.setDirection(QAbstractAnimation.Backward)
        self._opacity_anim.start()

    def _update_position(self):
        """
        Internal function that makes sure that panel is positioned in the proper place
        """

        parent = self.parent()
        parent_parent = parent.parent()
        dcc_win = tp.Dcc.get_main_window()
        dcc_window = parent_parent == dcc_win
        if parent_parent:
            dcc_window = dcc_window or parent_parent.objectName() == dcc_win.objectName()
        parent_geo = parent.geometry()
        if self._position == SliderPanelPositions.LEFT:
            pos = parent_geo.topLeft() if dcc_window else parent.mapToGlobal(parent_geo.topLeft())
            target_x = pos.x()
            target_y = pos.y()
            self.setFixedHeight(parent_geo.height())
            self._pos_anim.setStartValue(QPoint(target_x - self.width(), target_y))
            self._pos_anim.setEndValue(QPoint(target_x, target_y))
        if self._position == SliderPanelPositions.RIGHT:
            pos = parent_geo.topRight() if dcc_window else parent.mapToGlobal(parent_geo.topRight())
            self.setFixedHeight(parent_geo.height())
            target_x = pos.x() - self.width()
            target_y = pos.y()
            self._pos_anim.setStartValue(QPoint(target_x + self.width(), target_y))
            self._pos_anim.setEndValue(QPoint(target_x, target_y))
        if self._position == SliderPanelPositions.TOP:
            pos = parent_geo.topLeft() if dcc_window or parent_parent is None else parent.mapToGlobal(
                parent_geo.topLeft())
            self.setFixedWidth(parent_geo.width())
            target_x = pos.x()
            target_y = pos.y()
            self._pos_anim.setStartValue(QPoint(target_x, target_y - self.height()))
            self._pos_anim.setEndValue(QPoint(target_x, target_y))
        if self._position == SliderPanelPositions.BOTTOM:
            pos = parent_geo.bottomLeft() if dcc_window else parent.mapToGlobal(parent_geo.bottomLeft())
            self.setFixedWidth(parent_geo.width())
            target_x = pos.x()
            target_y = pos.y() - self.height()
            self._pos_anim.setStartValue(QPoint(target_x, target_y + self.height()))
            self._pos_anim.setEndValue(QPoint(target_x, target_y))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains custom Qt stack widgets
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import *
from Qt.QtWidgets import *

import tpDcc
from tpDcc.libs.qt.core import consts, qtutils, base, dpi, mixin
from tpDcc.libs.qt.widgets import layouts, buttons, lineedit


class SlidingStackedWidget(QStackedWidget, object):
    """
    QStackedWidget width sliding functionality
    """

    animFinished = Signal(int)

    def __init__(self, parent=None, **kwargs):
        super(SlidingStackedWidget, self).__init__(parent)

        # This variable holds the current widget without taking into account if the animation has finished or not
        self._speed = kwargs.pop('speed', 500)
        self._current_widget = None
        self._animation_type = kwargs.pop('animation_type', QEasingCurve.OutCubic)
        self._wrap = kwargs.pop('wrap', True)
        self._vertical = kwargs.pop('vertical', False)
        self._active_state = False
        self._blocked_page_list = list()
        self._now = 0
        self._next = 1

    @property
    def current_widget(self):
        return self._current_widget

    def set_speed(self, speed):
        """
        Sets the animation speed of the sliding
        :param speed: int
        """

        self._speed = speed

    def set_animation(self, animation_type):
        """
        Set the curve animation type of the sliding
        :param animation_type: QEasingCurve
        """

        self._animation_type = animation_type

    def set_vertical_mode(self, vertical=True):
        """
        Sets whether the sliding animation is done vertically or horizontally
        :param vertical: bool
        """

        self._vertical = vertical

    def set_wrap(self, wrap):
        """
        Sets whether the page index is restarted when we arrive the last one or not
        :param wrap: bool
        """

        self._wrap = wrap

    def slide_in_next(self):
        """
        Slides into the next widget
        """

        now = self.currentIndex()
        if self._wrap or now < self.count() - 1:
            self.slide_in_index(now + 1)

    def slide_in_prev(self):
        """
        Slides into the previous widget
        """

        now = self.currentIndex()
        if self._wrap or now > 0:
            self.slide_in_index(now - 1)

    def slide_in_index(self, next, force=False):
        """
        Slides to the given widget index
        :param next: int, index of the widget to slide
        """

        now = self.currentIndex()
        if (self._active_state or next == now) and not force:
            return

        self._active_state = True
        width, height = self.frameRect().width(), self.frameRect().height()
        next %= self.count()
        if next > now:
            if self._vertical:
                offset_x, offset_y = 0, height
            else:
                offset_x, offset_y = width, 0
        else:
            if self._vertical:
                offset_x, offset_y = 0, -height
            else:
                offset_x, offset_y = -width, 0
        self.widget(next).setGeometry(0, 0, width, height)
        pnow, pnext = self.widget(now).pos(), self.widget(next).pos()
        self._point_now = pnow

        self.widget(next).move(pnext.x() + offset_x, pnext.y() + offset_y)
        self.widget(next).show()
        self.widget(next).raise_()
        self._current_widget = self.widget(next)

        anim_now = QPropertyAnimation(self.widget(now), 'pos')
        anim_now.setDuration(self._speed)
        anim_now.setStartValue(pnow)
        anim_now.setEndValue(QPoint(pnow.x() - offset_x, pnow.y() - offset_y))
        anim_now.setEasingCurve(self._animation_type)

        anim_next = QPropertyAnimation(self.widget(next), 'pos')
        anim_next.setDuration(self._speed)
        anim_next.setStartValue(QPoint(offset_x + pnext.x(), offset_y + pnext.y()))
        anim_next.setEndValue(pnext)
        anim_next.setEasingCurve(self._animation_type)

        self._anim_group = QParallelAnimationGroup()
        self._anim_group.addAnimation(anim_now)
        self._anim_group.addAnimation(anim_next)
        self._anim_group.finished.connect(self._animation_done_slot)
        self._anim_group.start()

        self._next = next
        self._now = now

    def _animation_done_slot(self):
        self.setCurrentIndex(self._next)
        self.widget(self._now).hide()
        self.widget(self._now).move(self._point_now)
        try:
            self.widget(self._now).update()
        except Exception:
            pass
        self._active_state = False
        self.animFinished.emit(self._next)


@mixin.stacked_opacity_animation_mixin
class SlidingOpacityStackedWidget(QStackedWidget, object):
    """
    Custom stack widget that activates opacity animation when current stack index changes
    """

    def __init__(self, parent=None):
        super(SlidingOpacityStackedWidget, self).__init__(parent)


class StackItem(QFrame, object):

    minimized = Signal()
    maximized = Signal()
    toggleExpandRequested = Signal(bool)
    shiftUpPressed = Signal()
    shiftDownPressed = Signal()
    deletePressed = Signal()
    updateRequested = Signal()

    def __init__(self, title, parent, collapsed=False, collapsable=True, icon=None, start_hidden=False,
                 shift_arrows_enabled=True, delete_button_enabled=True, title_editable=True,
                 icon_size=12, title_frame=None, show_item_icon=True):
        super(StackItem, self).__init__(parent)

        self._stack_widget = parent
        self._icon_size = icon_size
        self._title = title
        self._title_editable = title_editable
        self._icon = icon
        self._shift_arrows_enabled = shift_arrows_enabled
        self._delete_button_enabled = delete_button_enabled
        self._color = consts.DARK_BG_COLOR
        self._contents_margins = (0, 0, 0, 0)
        self._contents_spacing = 0
        self._collapsable = collapsable
        self._collapsed = collapsed
        self._title_frame = title_frame
        self._show_item_icon = show_item_icon

        if start_hidden:
            self.hide()

        self.ui()
        self.setup_signals()

        if not collapsable:
            self._collapsed = False
            self.expand()

        if self._collapsed:
            self.collapse()
        else:
            self.expand()

    @property
    def contents_layout(self):
        return self._contents_layout

    @property
    def collapsed(self):
        return self._collapsed

    def ui(self):
        # theme = tpDcc.ToolsMgr().get_tool_theme('tpDcc-tools-hub')
        # border_width = qtutils.dpi_scale_divide(theme.STACK_BORDER_WIDTH)
        border_width = qtutils.dpi_scale_divide(1)
        self.main_layout = layouts.VerticalLayout(
            margins=(border_width, border_width, border_width, border_width), spacing=0)
        self.setLayout(self.main_layout)

        if not self._title_frame:
            self._title_frame = StackTitleFrame(
                title=self._title, icon=self._icon, title_editable=self._title_editable,
                item_icon_size=self._icon_size, collapsed=self._collapsed,
                shift_arrows_enabled=self._shift_arrows_enabled, delete_button_enabled=self._delete_button_enabled)
        if not self._show_item_icon:
            self._title_frame.item_icon_button.hide()

        self._widget_hider = StackHiderWidget(parent=self)
        self._contents_layout = layouts.VerticalLayout(spacing=0)
        self._contents_layout.setContentsMargins(*self._contents_margins)
        self._contents_layout.setSpacing(self._contents_spacing)
        self._widget_hider.setLayout(self._contents_layout)
        self._widget_hider.setContentsMargins(0, 0, 0, 0)
        self._widget_hider.setHidden(self._collapsed)
        self._widget_hider.setObjectName('stackbody')

        self.main_layout.addWidget(self._title_frame)
        self.main_layout.addWidget(self._widget_hider)

    def setup_signals(self):
        self._title_frame.expand_toggle_button.leftClicked.connect(self._on_toggle_contents)
        self._title_frame.minimized.connect(self.minimized.emit)
        self._title_frame.maximized.connect(self.maximized.emit)
        self._title_frame.toggleExpandRequested.connect(self.toggleExpandRequested.emit)
        self._title_frame.shiftUpPressed.connect(self.shiftUpPressed.emit)
        self._title_frame.shiftDownPressed.connect(self.shiftDownPressed.emit)
        self._title_frame.deletePressed.connect(self.deletePressed.emit)
        self._title_frame.updateRequested.connect(self.updateRequested.emit)

    def title_text_widget(self):
        """
        Returns title text widget
        :return: QLineEdit
        """

        return self._title_frame.line_edit

    def get_title(self):
        """
        Returns title text
        :return: str
        """

        return self._title_frame.line_edit.text()

    def set_title(self, text):
        """
        Function that sets title text
        :param text: str
        """

        self._title_frame.line_edit.setText(text)

    def add_widget(self, widget):
        """
        Adds a new widget to the contents layout
        :param widget: QWidget
        """

        self._contents_layout.addWidget(widget)

    def add_layout(self, layout):
        """
        Adds a new layout to the contents layout
        :param layout: QLayout
        """

        self._contents_layout.addLayout(layout)

    def expand(self, emit=True):
        """
        Function that expands the contents and show all the widget data
        :param emit: bool
        """

        self._on_expand(emit)

    def collapse(self, emit=True):
        """
        Function that collapses and hides the contents
        :param emit: bool
        """

        self._on_collapse(emit)

    def show_expand_indicator(self, flag):
        """
        Sets whether expand indicator is visible or not
        :param flag: bool
        """

        self._title_frame.expand_toggle_button.setVisible(flag)

    def set_title_text_mouse_transparent(self, flag):
        """
        Sets whether or not title text mouse is transparent
        :param flag: bool
        """

        self._title_frame.line_edit.setAttribute(Qt.WA_TransparentForMouseEvents, flag)

    def set_item_icon_color(self, color):
        """
        Sets the color of the item in title
        :param color: tuple(int, int, int), RGB color in 0-255 range
        """

        self._title_frame.set_item_icon_color(color)

    def set_item_icon(self, icon):
        """
        Sets the icon of the item in title
        :param icon: QIcon
        """

        self.title_frame.set_item_icon(icon)

    def shift_up(self):
        """
        Emits item shift up signal
        """

        self.shiftUpPressed.emit()

    def shift_down(self):
        """
        Emits shijft down signal
        """

        self.shiftDownPressed.emit()

    def set_arrows_visible(self, flag):
        """
        Sets whether shift arrows are visible or not.
        This arrows allow the item to be shifted upwards or downwards the stack
        :param flag: bool
        """

        if flag:
            self._title_frame.shift_down_button.show()
            self._title_frame.shift_up_button.show()
        else:
            self._title_frame.shift_down_button.hide()
            self._title_frame.shift_up_button.hide()

    def update_size(self):
        """
        Updates the size of the widget, to fit to new size depending on its contents
        """

        self.updateRequested.emit()

    def _on_expand(self, emit=True):
        """
        Internal function that expands the contents and show all the widget data
        :param emit: bool
        """

        self._widget_hider.setHidden(False)
        self._title_frame.expand()
        if emit:
            self.maximized.emit()
        self._collapsed = False

    def _on_collapse(self, emit=True):
        """
        Internal function that collapses and hides the contents
        :param emit: bool
        """

        self._widget_hider.setHidden(True)
        self._title_frame.collapse()
        if emit:
            self.minimized.emit()
        self._collapsed = True

    def _on_toggle_contents(self, emit=True):
        """
        Internal callback function that shows and hides the hidder widget and shows/hides contents
        :param emit: bool
        """

        if not self._collapsable:
            return

        self.toggleExpandRequested.emit(not self._collapsed)
        if self._collapsed:
            self.expand(emit)
            self.update_size()
            return not self._collapsed

        self.collapse(emit)
        self.update_size()
        return self._collapsed


class StackHiderWidget(base.BaseFrame, object):
    def __init__(self, parent=None):
        super(StackHiderWidget, self).__init__(parent)


class StackTitleFrame(QFrame, dpi.DPIScaling):

    minimized = Signal()
    maximized = Signal()
    toggleExpandRequested = Signal(bool)
    shiftUpPressed = Signal()
    shiftDownPressed = Signal()
    deletePressed = Signal()
    updateRequested = Signal()

    def __init__(self, title='', title_editable=False, icon=None, item_icon_size=16, collapsed=True,
                 shift_arrows_enabled=True, delete_button_enabled=True, parent=None):
        super(StackTitleFrame, self).__init__(parent)

        self._title = title
        self._item_icon = icon or 'tpdcc'
        self._title_editable = title_editable
        self._shift_arrows_enabled = shift_arrows_enabled
        self._delete_button_enabled = delete_button_enabled
        self._spaces_to_underscore = False
        self._collapsed = collapsed
        self._item_icon_size = item_icon_size

        self._collapsed_icon = tpDcc.ResourcesMgr().icon('sort_closed')
        self._expand_icon = tpDcc.ResourcesMgr().icon('sort_down')

        self.setObjectName('title')

        self.ui()
        self.setup_signals()

        self.setFixedHeight(self.sizeHint().height())
        self.setMinimumSize(self.sizeHint().width(), self.sizeHint().height() + 1)

    @property
    def horizontal_layout(self):
        return self._horizontal_layout

    @property
    def item_icon(self):
        return self._item_icon

    @property
    def item_icon_button(self):
        return self._item_icon_btn

    @property
    def expand_toggle_button(self):
        return self._expand_toggle_btn

    @property
    def shift_down_button(self):
        return self._shift_down_btn

    @property
    def shift_up_button(self):
        return self._shift_up_btn

    @property
    def delete_button(self):
        return self._delete_btn

    @property
    def line_edit(self):
        return self._line_edit

    def mouseDoubleClickEvent(self, event):
        if self._title_editable:
            self._line_edit.editEvent(event)

    def ui(self):

        item_icon = tpDcc.ResourcesMgr().icon(self._item_icon)
        if item_icon.isNull():
            item_icon = tpDcc.ResourcesMgr().icon('tpdcc')
        delete_icon = tpDcc.ResourcesMgr().icon('delete')
        shift_up_icon = tpDcc.ResourcesMgr().icon('arrow_up')
        shift_down_icon = tpDcc.ResourcesMgr().icon('arrow_down')

        self.setContentsMargins(*qtutils.margins_dpi_scale(0, 0, 0, 0))
        self._extras_layout = layouts.HorizontalLayout(margins=(0, 0, 0, 0), spacing=0)
        self._horizontal_layout = layouts.HorizontalLayout(spacing=0)
        self.setLayout(self._horizontal_layout)

        self._line_edit = lineedit.ClickLineEdit(self._title)
        self._line_edit.setObjectName('lineEdit')
        self._line_edit.setFocusPolicy(Qt.NoFocus)
        self._line_edit.setVisible(False)
        self._line_edit.setAttribute(Qt.WA_TransparentForMouseEvents)
        if not self._title_editable:
            self._line_edit.setReadOnly(True)

        self._expand_toggle_btn = buttons.BaseMenuButton()
        self._item_icon_btn = buttons.BaseMenuButton()
        self._shift_down_btn = buttons.BaseMenuButton()
        self._shift_up_btn = buttons.BaseMenuButton()
        self._delete_btn = buttons.BaseMenuButton()

        if not self._shift_arrows_enabled:
            self._shift_down_btn.hide()
            self._shift_up_btn.hide()
        if not self._delete_button_enabled:
            self._delete_btn.hide()

        self._item_icon_btn.set_icon(icon=item_icon, size=self._item_icon_size)
        self._item_icon_btn.setAttribute(Qt.WA_TransparentForMouseEvents)

        icon_size = 12
        highlight_offset = 40
        self._delete_btn.set_icon(delete_icon, colors=None, size=icon_size, color_offset=highlight_offset)
        self._shift_up_btn.set_icon(shift_up_icon, colors=None, size=icon_size, color_offset=highlight_offset)
        self._shift_down_btn.set_icon(shift_down_icon, colors=None, size=icon_size, color_offset=highlight_offset)
        self._expand_toggle_btn.set_icon(self._expand_icon, colors=(192, 192, 192), size=icon_size)

        self._horizontal_layout.addWidget(self._item_icon_btn)
        self._horizontal_layout.addItem(QSpacerItem(10, 10, QSizePolicy.Expanding, QSizePolicy.Minimum))
        self._horizontal_layout.addWidget(self._line_edit)
        self._horizontal_layout.addLayout(self._extras_layout)
        self._horizontal_layout.addWidget(self._shift_up_btn)
        self._horizontal_layout.addWidget(self._shift_down_btn)
        self._horizontal_layout.addWidget(self._expand_toggle_btn)
        self._horizontal_layout.addWidget(self._delete_btn)
        self._horizontal_layout.setStretchFactor(self._line_edit, 4)

    def setup_signals(self):
        self._shift_up_btn.leftClicked.connect(self._on_shift_up)
        self._shift_down_btn.leftClicked.connect(self._on_shift_down)
        self._delete_btn.leftClicked.connect(self.deletePressed.emit)
        self._line_edit.textChanged.connect(self._on_title_validate)
        self._line_edit.selectionChanged.connect(self._on_select_check)

    def collapse(self):
        self._expand_toggle_btn.set_icon(self._collapsed_icon, colors=(192, 192, 192), size=12)

    def expand(self):
        self._expand_toggle_btn.set_icon(self._expand_icon, colors=(192, 192, 192), size=12)

    def set_item_icon_color(self, color):
        self._item_icon_btn.set_icon_color(color)

    def set_item_icon(self, icon):
        self._item_icon_btn.set_icon(icon)

    def _on_shift_up(self):
        self.shiftUpPressed.emit()

    def _on_shift_down(self):
        self.shiftDownPressed.emit()

    def _on_title_validate(self):
        """
        Internal callback function that is called when line edit text changes
        Removes invalid characters and replaces spaces with underscores
        """

        if self._spaces_to_underscore:
            line_edit = self._line_edit
            text = self._line_edit.text()
            pos = line_edit.cursorPosition()
            text = text.replace(' ', '_')
            line_edit.blockSignals(True)
            try:
                line_edit.setText(text)
            finally:
                line_edit.blockSignals(False)
            line_edit.setCursorPosition(pos)

    def _on_select_check(self):
        """
        Internal callback function that is called when user deselects line edit
        Stops any selection from happening if title is not editable
        """

        if not self._title_editable:
            self._line_edit.deselect()
